"""PVT implements a class to describe (position, velocity, time) triplets.

History:
2001-01-10 ROwen    Modified floatCnv to not handle NaN floating values,
    since this failed on Mac OS X; it will still handle string "NaN" (any case).
2002-08-08 ROwen    Modified to use new Astro.Tm functions which are in days, not sec.
2003-05-08 ROwen    Modified to use RO.CnvUtil.
2003-11-21 ROwen    Bug fix: __init__ did not check the data.
2005-06-08 ROwen    Changed PVT to a new-style class.
2007-07-02 ROwen    Added hasVel method.
2015-09-24 ROwen    Replace "== None" with "is None" to modernize the code.
2015-11-03 ROwen    Replace "!= None" with "is not None" to modernize the code.
"""

import math
import time


__all__ = ['PVT']


MJDJ2000 = 51544.5      # Modified Julian Date at epoch J2000.0 noon (days)
SecPerDay = 24.0 * 3600.0  # seconds per day
_UTCMinusTAIDays = -35 / float(SecPerDay)  # a reasonable value correct as of 2012-12
_TimeError = 0.0  # time reported by your computer's clock - actual time (seconds)


def getCurrPySec(uncorrTime=None):
    """Get current python time with time error correction applied

    Input:
    - uncorrTime: python time without correction applied; if None then current time is used

    """
    if uncorrTime is None:
        uncorrTime = time.time()
    return uncorrTime - _TimeError


def utcFromPySec(pySec=None):
    """Returns the UTC (MJD) corresponding to the supplied python time, or now if none.
    """
    global _TimeTupleJ2000

    if pySec is None:
        pySec = getCurrPySec()

    # python time (in seconds) corresponding to 2000-01-01 00:00:00
    # this is probably constant, but there's some chance
    # that on some computer systems it varies with daylights savings time
    pySecJ2000 = time.mktime(_TimeTupleJ2000) - time.timezone

    return MJDJ2000 + ((pySec - pySecJ2000) / SecPerDay)


def taiFromUTC(utc):
    """Convert UTC (MJD) to TAI (MJD)"""

    return utc - _UTCMinusTAIDays


def taiFromPySec(pySec=None):
    """Convert python seconds (now if None) to TAI (MJD)"""

    return taiFromUTC(utcFromPySec(pySec))


def asFloatOrNone(val):
    """Converts floats, integers and string representations of either to floats.
    If val is "NaN" (case irrelevant) or "?" returns None.

    Raises ValueError or TypeError for all other values

    """

    # check for NaN first in case ieee floating point is in use
    # (in which case float(val) would return something instead of failing)

    if hasattr(val, 'lower') and val.lower() in ('nan', '?'):
        return None
    else:
        return float(val)


class PVT(object):
    """Defines a position, velocity, time triplet, where time is in TAI.

    Inputs:
    - pos   position
    - vel   velocity (in units of position/sec)
    - time  TAI, MJD seconds

    Each value must be one of: a float, a string representation of a float,
    "NaN" (any case) or None. "NaN" and None mean "unknown" and are stored as None.

    Raises ValueError if any value is invalid.
    """

    def __init__(self, pos=None, vel=0.0, t=0.0):
        self.pos = None
        self.vel = 0.0
        self.t = 0.0
        self.set(pos, vel, t)

    def __repr__(self):
        return 'PVT(%s, %s, %s)' % (str(self.pos), str(self.vel), str(self.t))

    @property
    def native(self):
        """Returns a tuple of ``(pos, vel, t)``.

        For compatibility with other opscore types.

        """

        return (self.pos, self.vel, self.t)

    def getPos(self, t=None):
        """Returns the position at the specified time.
        Time defaults to the current TAI.

        Returns None if the pvt is invalid.
        """
        if not self.isValid():
            return None

        if t is None:
            t = taiFromPySec() * SecPerDay

        return self.pos + (self.vel * (t - self.t))

    def hasVel(self):
        """Return True if velocity is known and nonzero.
        """
        return self.vel not in (0, None)

    def isValid(self):
        """Returns True if the pvt is valid, False otherwise.

        A pvt is valid if all values are known (not None and finite) and time > 0.
        """
        return ((self.pos is not None) and math.isfinite(self.pos) and
                (self.vel is not None) and math.isfinite(self.vel) and
                (self.t is not None) and math.isfinite(self.t) and
                (self.t > 0))

    def set(self, pos=None, vel=None, t=None):
        """Sets pos, vel and t; all default to their current values

        Each value must be one of: a float, a string representation of a float,
        "NaN" (any case) or None. "NaN" means "unknown" and is stored as None.

        Errors:
        Raises ValueError if any value is invalid.
        """
        if pos is not None:
            self.pos = asFloatOrNone(pos)
        if vel is not None:
            self.vel = asFloatOrNone(vel)
        if t is not None:
            self.t = asFloatOrNone(t)


if __name__ == '__main__':
    print('\nrunning PVT test')

    currTAI = taiFromPySec() * SecPerDay

    varList = (PVT(), PVT(25), PVT(25, 0, currTAI), PVT(25, 1), PVT(25, 1, currTAI),
               PVT('NaN', 'NaN', 'NaN'))

    for i in range(5):
        t = taiFromPySec() * SecPerDay
        print('\ntime =', t)
        for var in varList:
            print(var, 'pos =', var.getPos(t))
        if i < 4:
            time.sleep(1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2019-05-13
# @Filename: actor.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import warnings

import clu

from ..actor import TimerCommandList
from ..base import BaseActor
from ..command import Command, parse_legacy_command
from ..parser import ClickParser
from ..protocol import TCPStreamServer
from ..tools import log_reply
from .tron import TronConnection


__all__ = ['LegacyActor', 'BaseLegacyActor']


class BaseLegacyActor(BaseActor):
    """An actor that provides compatibility with the SDSS opscore protocol.

    The TCP servers need to be started by awaiting the coroutine `.start`.
    Note that `.start` does not block so you will need to use asyncio's
    `.run_forever` or a similar system ::

        >>> loop = asyncio.get_event_loop()
        >>> my_actor = await LegacyActor('my_actor', '127.0.0.1', 9999, loop=loop).start()
        >>> loop.run_forever()

    Parameters
    ----------
    name : str
        The name of the actor.
    host : str
        The host where the TCP server will run.
    port : int
        The port of the TCP server.
    tron_host : str
        The host on which Tron is running.
    tron_port : int
        The port on which Tron is running.
    model_names : list
        A list of strings with the actors whose models will be tracked.
    version : str
        The version of the actor.
    loop
        The event loop. If `None`, the current event loop will be used.
    log_dir : str
        The directory where to store the logs. Defaults to
        ``$HOME/logs/<name>`` where ``<name>`` is the name of the actor.
    log : ~logging.Logger
        A `~logging.Logger` instance to be used for logging instead of creating
        a new one.

    """

    host = None
    port = None

    def __init__(self, name, host, port, tron_host=None, tron_port=None,
                 model_names=None, version=None, loop=None, log_dir=None,
                 log=None):

        super().__init__(name, version=version, loop=loop,
                         log_dir=log_dir, log=log)

        #: Mapping of user_id to transport
        self.transports = dict()

        self.host = host
        self.port = port

        #: TCPStreamServer: The server to talk to this actor.
        self.server = TCPStreamServer(host, port, loop=self.loop,
                                      connection_callback=self.new_user,
                                      data_received_callback=self.new_command)

        if tron_host and tron_port:
            #: TronConnection: The client connection to Tron.
            self.tron = TronConnection(self.name, tron_host, tron_port,
                                       model_names=model_names, log=self.log)
        else:
            self.tron = False

        if self.tron:
            #: dict: Actor models.
            self.models = self.tron.models

        self.timer_commands = TimerCommandList(self)

    def __repr__(self):

        return f'<{str(self)} (name={self.name!r}, host={self.host!r}, port={self.port})>'

    async def start(self):
        """Starts the server and the Tron client connection."""

        await self.server.start_server()
        self.log.info(f'running TCP server on {self.host}:{self.port}')

        # Start tron connection
        try:
            if self.tron:
                await self.tron.start()
                self.log.info('started tron connection at '
                              f'{self.tron.host}:{self.tron.port}')
            else:
                warnings.warn('starting LegacyActor without Tron connection.',
                              clu.CluWarning)
        except (ConnectionRefusedError, OSError) as ee:
            warnings.warn(f'connection to tron was refused: {ee}. '
                          'Some functionality will be limited.', clu.CluWarning)

        self.timer_commands.start()

        return self

    async def run_forever(self):
        """Runs the actor forever, keeping the loop alive."""

        await self.server.serve_forever()

    def new_user(self, transport):
        """Assigns userID to new client connection."""

        if transport.is_closing():
            if hasattr(transport, 'user_id'):
                self.log.debug(f'user {transport.user_id} disconnected.')
                return self.transports.pop(transport.user_id)

        curr_ids = set(self.transports.keys())
        user_id = 1 if len(curr_ids) == 0 else max(curr_ids) + 1

        transport.user_id = user_id

        self.transports[user_id] = transport

        # report user information and additional info
        self.show_new_user_info(user_id)

        return

    def new_command(self, transport, command_str):
        """Handles a new command received by the actor."""

        commander_id = getattr(transport, 'user_id', 0)
        command_str = command_str.decode().strip()

        if not command_str:
            return
        try:
            command_id, command_body = parse_legacy_command(command_str)

            command = Command(command_string=command_body, commander_id=commander_id,
                              command_id=command_id, consumer_id=self.name,
                              actor=self, loop=self.loop, transport=transport)
        except clu.CommandError as ee:
            self.write('f', {'text': f'Could not parse the following as a command: {ee!r}'})
            return

        return self.parse_command(command)

    @staticmethod
    def format_user_output(user_id, command_id, message_code, msg_str=None):
        """Formats a string to send to users."""

        msg_str = '' if msg_str is None else ' ' + msg_str

        return f'{user_id} {command_id:d} {message_code:s}{msg_str:s}'

    def show_new_user_info(self, user_id):
        """Shows information for new users. Called when a new user connects."""

        self.show_user_info(user_id)
        self.show_version(user_id=user_id)

        transport = self.transports[user_id]
        peername = transport.get_extra_info('peername')[0]
        self.log.debug(f'user {user_id} connected from {peername!r}.')

    def show_user_info(self, user_id):
        """Shows user information including your user_id."""

        num_users = len(self.transports)
        if num_users == 0:
            return

        msg = {'yourUserID': user_id,
               'num_users': num_users}

        self.write('i', msg, user_id=user_id)
        self.show_user_list()

    def show_user_list(self):
        """Shows a list of connected users. Broadcast to all users."""

        user_id_list = sorted(self.transports.keys())
        for user_id in user_id_list:
            transport = self.transports[user_id]
            peername = transport.get_extra_info('peername')[0]
            msg = {'UserInfo': f'{user_id}, {peername}'}
            self.write('i', msg)

    @staticmethod
    def get_user_command_id(command=None, user_id=None, command_id=None):
        """Returns commander_id, command_id based on user-supplied information.

        Parameters
        ----------
        command : Command
            User command; used as a default for ``user_id`` and
            ``command_id``.
        user_id : int
            If `None` then use ``command.user_id``.
        command_id : int
            If `None` then use ``command.command_id``.

        Returns
        -------
        user_id, command_id : `tuple`
            The commander ID and the command ID, parsed from the inputs. If
            they cannot be determined, returns zeros.

        """

        user_id = user_id or (command.commander_id if command else 0)
        command_id = command_id or (command.command_id if command else 0)

        return (user_id, command_id)

    def show_version(self, user_id=None):
        """Shows actor version."""

        msg = {'version': repr(self.version)}

        self.write('i', msg, user_id=user_id)

    def send_command(self, target, command_string, command_id=None):
        """Sends a command through the hub.

        Parameters
        ----------
        target : str
            The actor to command.
        command_string : str
            The command to send.
        command_id : int
            The command id. If `None`, a sequentially increasing value will
            be used. You should not specify a ``command_id`` unless you really
            know what you're doing.

        """

        if self.tron.connection:
            self.tron.send_command(target, command_string, mid=command_id)
        else:
            raise clu.CluError('cannot connect to tron.')

    def write(self, message_code='i', message=None, command=None, user_id=None,
              command_id=None, concatenate=True, broadcast=False, **kwargs):
        """Writes a message to user(s).

        Parameters
        ----------
        message_code : str
            The message code (e.g., ``'i'`` or ``':'``).
        message : dict
            The keywords to be output. Must be a dictionary of pairs
            ``{keyword: value}``. If ``value`` is a list it will be converted
            into a comma-separated string. To prevent unexpected casting,
            it is recommended for ``value`` to always be a string.
        command : Command
            User command; used as a default for ``user_id`` and
            ``command_id``. If the command is done, it is ignored.
        user_id : int
            The user (transport) to which to write. `None` defaults to 0.
        command_id : int
            If `None` then use ``command.command_id``.
        concatenate : bool
            Concatenates all the keywords to be output in a single
            reply with the keyword-values joined with semicolons. Otherwise
            each keyword will be output as a different message.
        broadcast : bool
            Whether to broadcast the reply. Equivalent to ``user_id=0``.
        kwargs
            Keyword arguments that will be added to the message. If a keyword
            is both in ``message`` and in ``kwargs``, the value in ``kwargs``
            supersedes ``message``.

        """

        # For a reply, the commander ID is the user assigned to the transport
        # that issues this command.
        transport = command.transport if command else None
        user_id = transport.user_id if transport else 0

        if broadcast:
            user_id = 0
            command_id = 0

        user_id, command_id = self.get_user_command_id(
            command=command, user_id=user_id, command_id=command_id)

        if message is None or (isinstance(message, str) and message.strip() == ''):
            message = {}
        elif isinstance(message, str):
            message = {'text': message}
        elif not isinstance(message, dict):
            raise TypeError('invalid message type ' + str(type(message)))

        message.update(kwargs)

        lines = []
        for keyword in message:
            value = clu.format_value(message[keyword])
            lines.append(f'{keyword}={value}')

        if concatenate:
            lines = ['; '.join(lines)]

        for line in lines:

            full_msg_str = self.format_user_output(user_id, command_id,
                                                   message_code, line)
            msg = (full_msg_str + '\n').encode()

            if user_id is None or user_id == 0 or transport is None:
                for transport in self.transports.values():
                    transport.write(msg)
            else:
                transport.write(msg)

            log_reply(self.log, message_code, full_msg_str)


class LegacyActor(ClickParser, BaseLegacyActor):
    """A legacy actor that uses the `.ClickParser`."""



"""
Type declarations for SDSS-3 keyword values

Refer to https://trac.sdss3.org/wiki/Ops/Types
"""

# Created 30-Oct-2008 by David Kirkby (dkirkby@uci.edu)

import re
import textwrap

from . import html, pvt


class ValueTypeError(Exception):
    pass


class InvalidValueError(Exception):
    """
    Signals that an invalid value was detected during conversion from a string
    """
    pass


class Descriptive(object):
    """
    A self-describing type mixin

    Types that use this mixin are responsible for filling an array cls.descriptors
    of tuples (label,value) in their constructor.
    """

    def describe(self):
        """
        Returns a plain-text multi-line description

        The final newline character is omitted so that the return value
        can be printed. The description is indented with spaces assuming
        a fixed-width font.
        """
        text = ''
        for label, value in self.descriptors:
            pad = '\n' + ' ' * 14
            formatted = textwrap.fill(textwrap.dedent(value).strip(), width=66).replace('\n', pad)
            text += '%12s: %s\n' % (label, formatted)
        return text[:-1]

    def describeAsHTML(self):
        """
        Returns a description packaged as an HTML fragment

        Uses the following CSS elements: div.vtype, div.descriptor,
        span.label, span.value
        """
        content = html.Div(className='vtype')
        for label, value in self.descriptors:
            content.append(
                html.Div(
                    html.Span(label, className='label'),
                    html.Span(value, className='value'),
                    className='type descriptor'))
        return content


class ValueType(type, Descriptive):
    """
    A metaclass for types that represent an enumerated or numeric value
    """
    _nameSpec = re.compile('[A-Za-z][A-Za-z0-9_]*')
    _metaKeys = ('reprFmt', 'strFmt', 'invalid', 'units', 'help', 'name')

    def __new__(cls, *args, **kwargs):
        """
        Allocates memory for a new ValueType class
        """

        def doRepr(self):
            if self.units:
                units = ' ' + self.units
            else:
                units = ''
            if self.reprFmt:
                return '%s(%s%s)' % (cls.__name__, self.reprFmt % self, units)
            else:
                return '%s(%s%s)' % (cls.__name__, cls.baseType.__repr__(self), units)

        def doStr(self):
            if self.strFmt:
                return self.strFmt % self
            elif self.reprFmt:
                return self.reprFmt % self
            else:
                return cls.baseType.__str__(self)

        # check for any invalid metadata keys
        for key in kwargs:
            if key not in ValueType._metaKeys and (not hasattr(cls, 'customKeys') or
                                                   key not in cls.customKeys):
                raise ValueTypeError('invalid metadata key "%s" for %s' % (key, cls.__name__))

        # force the invalid string, if present, to be lowercase
        if 'invalid' in kwargs:
            kwargs['invalid'] = str(kwargs['invalid']).lower()

        # check that the name string, if present, is a valid identifier
        if 'name' in kwargs:
            matched = ValueType._nameSpec.match(kwargs['name'])
            if not matched or not matched.end() == len(kwargs['name']):
                raise ValueTypeError('invalid type name: %s' % kwargs['name'])

        # check that any format strings provided are not circular
        if 'reprFmt' in kwargs:
            fmt = kwargs['reprFmt']
            if (fmt.find('%s') != -1) or (fmt.find('%r') != -1):
                raise ValueTypeError('reprFmt cannot contain %r or %s')
        if 'strFmt' in kwargs:
            fmt = kwargs['strFmt']
            if (fmt.find('%s') != -1) or (fmt.find('%r') != -1):
                raise ValueTypeError('strFmt cannot contain %r or %s')

        def get(name, default=None):
            return kwargs.get(name, cls.__dict__.get(name, default))

        dct = {
            'reprFmt': get('reprFmt'),
            'strFmt': get('strFmt'),
            'invalid': get('invalid'),
            'units': get('units'),
            'help': get('help'),
            'name': get('name'),
            '__repr__': doRepr,
            '__str__': doStr,
        }
        if cls == Bits:
            # leave special bit handling alone, since some code may rely on it
            def getNative(self):
                return int(self)

            dct['native'] = property(getNative)
        elif cls == Bool:
            # cls.baseType is int, not bool, because one cannot subclass bool
            def getNative(self):
                return bool(self)

            dct['native'] = property(getNative)
        elif hasattr(cls, 'baseType'):

            def getNative(self):
                return cls.baseType(self)

            dct['native'] = property(getNative)
        else:
            print('WARNING: no baseType')

        if hasattr(cls, 'new'):
            dct['__new__'] = cls.new
        if hasattr(cls, 'init'):
            cls.init(dct, *args, **kwargs)
        return type.__new__(cls, cls.__name__, (cls.baseType, ), dct)

    def addDescriptors(cls):
        if cls.units:
            cls.descriptors.append(('Units', cls.units))

    def __init__(cls, *args, **kwargs):
        """
        Initializes a new ValueType class
        """
        super(ValueType, cls).__init__(cls.__name__, (cls.baseType, ), {})
        cls.descriptors = []
        if cls.name:
            cls.descriptors.append(('Name', cls.name))
        if cls.help:
            cls.descriptors.append(('Description', cls.help))
        cls.descriptors.append(('Type',
                                '%s (%s,%s)' % (cls.__name__, cls.baseType.__name__, cls.storage)))
        cls.addDescriptors()
        if cls.invalid:
            cls.descriptors.append(('Invalid', cls.invalid))

    def __mul__(self, amount):
        if isinstance(amount, int):
            return RepeatedValueType(self, amount, amount)
        elif not isinstance(amount, tuple):
            raise TypeError('Cannot multiply ValueType by type %r' % type(amount))
        if len(amount) == 0 or len(amount) > 2:
            raise ValueTypeError('Repetions should be specified as *(min,) or *(min,max)')
        minRepeat = amount[0]
        if len(amount) == 2:
            maxRepeat = amount[1]
        else:
            maxRepeat = None
        return RepeatedValueType(self, minRepeat, maxRepeat)

    def __repr__(self):
        return self.__name__

    def validate(self, value):
        if self.invalid and str(value).lower() == self.invalid:
            raise InvalidValueError
        return value


class RepeatedValueType(Descriptive):

    def __init__(self, vtype, minRepeat, maxRepeat):
        if not isinstance(vtype, ValueType):
            raise ValueTypeError('RepeatedValueType only works for a ValueType')
        self.vtype = vtype
        if not isinstance(minRepeat, int) or (maxRepeat and not isinstance(maxRepeat, int)):
            raise ValueTypeError('Expected integer min/max repetitions for RepeatedValueType')
        self.minRepeat = minRepeat
        self.maxRepeat = maxRepeat
        if self.minRepeat < 0 or (self.maxRepeat and self.maxRepeat < self.minRepeat):
            raise ValueTypeError('Expected min <= max for RepeatedValueType')
        if self.minRepeat == 1:
            times = 'once'
        else:
            times = '%d times' % self.minRepeat
        if self.minRepeat == self.maxRepeat:
            repeatText = times
        elif self.maxRepeat is None:
            repeatText = 'at least ' + times
        else:
            repeatText = '%d-%d times' % (self.minRepeat, self.maxRepeat)
        self.descriptors = [('Repeated', repeatText)]
        self.descriptors.extend(self.vtype.descriptors)

    def __repr__(self):
        if self.minRepeat == self.maxRepeat:
            return '%r*%d' % (self.vtype, self.minRepeat)
        elif self.maxRepeat is None:
            return '%r*(%d,)' % (self.vtype, self.minRepeat)
        else:
            return '%r*(%d,%d)' % (self.vtype, self.minRepeat, self.maxRepeat)


class CompoundValueType(Descriptive):
    """
    Represents a compound type consisting of a sequence of simple types

    A compound value is normally represented by a single object. By default,
    the wrapping object is a tuple containing the individual values but a
    custom object can be used via the 'wrapper' keyword which should provide
    a function (which might be a class constructor) that is called with the
    individual values and returns the wrapping object.
    """
    # a global flag to enable/disable the wrapping of compound values
    WrapEnable = True

    def __init__(self, *vtypes, **kwargs):
        self.vtypes = vtypes
        self.name = kwargs.get('name', None)
        self.help = kwargs.get('help', None)
        self.descriptors = []
        if self.name:
            self.descriptors.append(('Name', self.name))
        if self.help:
            self.descriptors.append(('Description', self.help))
        for index, vtype in enumerate(self.vtypes):
            self.descriptors.append(('Subtype-%d' % index, '-' * 40))
            self.descriptors.extend(vtype.descriptors)
        self.wrapper = kwargs.get('wrapper', None)

    def __repr__(self):
        return '%s%r' % (self.__class__.__name__, self.vtypes)


class PVT(CompoundValueType):
    """
    Represents a position-velocity-time compound type
    """

    def __init__(self, **kwargs):
        vtypes = (Float(name='position', units='deg'), Float(name='velocity', units='deg/s'),
                  Double(name='time', units='MJD-secs(TAI)'))
        if 'wrapper' not in kwargs:
            # by default, use RO.PVT to wrap values of type PVT

            kwargs['wrapper'] = pvt.PVT
        CompoundValueType.__init__(self, *vtypes, **kwargs)


class Invalid(object):
    """
    Represents an invalid value
    """
    units = ''
    native = None

    def __repr__(self):
        return '(invalid)'

    def __eq__(self, other):
        """
        All Invalid instances are equal to each other and to None
        """
        return isinstance(other, Invalid) or other is None

    def __ne__(self, other):
        return not self.__eq__(other)


# a constant object representing an invalid value
InvalidValue = Invalid()


class Float(ValueType):
    baseType = float
    storage = 'flt4'

    def new(cls, value):
        fvalue = float(cls.validate(value))
        # the limit value is float(340282346638528859811704183484516925440)
        # where 3402... is (2 - 2^(-23)) 2^127
        if abs(fvalue) > 3.4028234663852886e+38 and abs(fvalue) != float('inf'):
            raise OverflowError('Invalid literal for Float: %r' % value)
        return float.__new__(cls, fvalue)


class Double(ValueType):
    baseType = float
    storage = 'flt8'

    def new(cls, value):
        return float.__new__(cls, cls.validate(value))


class Int(ValueType):
    baseType = int
    storage = 'int4'

    def new(cls, value):
        if isinstance(value, str):
            # base = 0 infers base from optional prefix (see PEP 3127)
            lvalue = int(cls.validate(value), 0)
        else:
            lvalue = int(cls.validate(value))
        if lvalue < -0x7fffffff or lvalue > 0x7fffffff:
            raise OverflowError('Invalid literal for Int: %r' % value)
        return int.__new__(cls, lvalue)


class Long(ValueType):
    baseType = int
    storage = 'int8'

    def new(cls, value):
        if isinstance(value, str):
            # base = 0 infers base from optional prefix (see PEP 3127)
            return int.__new__(cls, cls.validate(value), 0)
        else:
            return int.__new__(cls, cls.validate(value))


class String(ValueType):
    baseType = str
    storage = 'text'

    def new(cls, value):
        return str.__new__(cls, cls.validate(value))


class UInt(ValueType):
    baseType = int
    storage = 'int4'

    def new(cls, value):
        if isinstance(value, str):
            # base = 0 infers base from optional prefix (see PEP 3127)
            lvalue = int(cls.validate(value), 0)
        else:
            lvalue = int(cls.validate(value))
        if lvalue < -0x7fffffff or lvalue > 0xffffffff:
            raise OverflowError('Invalid literal for UInt: %r' % value)
        if lvalue < 0:
            # re-interpret a negative 32-bit value as its bit-equivalent unsigned value
            lvalue = 0x80000000 | (-lvalue)
        return int.__new__(cls, lvalue)


class Hex(UInt):
    """
    The Hex class has been deprecated and is scheduled for deletion (20-Jul-2009)
    """
    reprFmt = '0x%x'

    def new(cls, value):
        try:
            return int.__new__(cls, cls.validate(value), 16)
        except TypeError:
            return UInt.new(cls, value)


# Enumerated value type
class Enum(ValueType):

    baseType = str
    storage = 'int2'
    customKeys = ('labelHelp')

    @classmethod
    def init(cls, dct, *args, **kwargs):
        if not args:
            raise ValueTypeError('missing enum labels in ctor')
        # force each label to be interpreted as a string so, for example,
        # False->'False', 1->'1', 0xff->'255'
        strargs = [str(arg) for arg in args]
        dct['enumLabels'] = strargs
        dct['enumValues'] = dict(zip(args, range(len(strargs))))
        # look for optional per-label help text
        labelHelp = kwargs.get('labelHelp', None)
        if labelHelp and not len(labelHelp) == len(strargs):
            raise ValueTypeError('wrong number of enum label help strings provided')
        dct['labelHelp'] = labelHelp

        # provide a custom storage value helper since our storage type is int2
        # but our basetype is str
        def storageValue(self):
            return str(self.enumValues[self])

        dct['storageValue'] = storageValue

        # enumerated value comparisons are case insensitive
        def eqTest(self, other):
            return str(other).lower() == self.lower()

        def neTest(self, other):
            return str(other).lower() != self.lower()

        dct['__eq__'] = eqTest
        dct['__ne__'] = neTest

    def new(cls, value):
        """
        Initializes a new enumerated instance

        Value can either be an integer index or else a recognized label.
        """
        cls.validate(value)
        if isinstance(value, int):
            if value >= 0 and value < len(cls.enumLabels):
                return str.__new__(cls, cls.enumLabels[value])
            else:
                raise ValueError('Invalid index for Enum: %d' % value)
        value = str(value).lower()
        for label in cls.enumLabels:
            if value == label.lower():
                return str.__new__(cls, label)
        raise ValueError('Invalid label for Enum: "%s"' % value)

    def addDescriptors(cls):
        for index, label in enumerate(cls.enumLabels):
            description = label
            if cls.labelHelp:
                description += ' (%s)' % cls.labelHelp[index]
            cls.descriptors.append(('Value-%d' % index, description))


# Boolean value type
class Bool(ValueType):

    baseType = int  # bool cannot be subclassed
    storage = 'int2'

    @classmethod
    def init(cls, dct, *args, **kwargs):
        if not args or not len(args) == 2:
            raise ValueTypeError('missing true/false labels in ctor')
        # force the literal values to be interpreted as strings so, for example,
        # False->'False', 0->'0'
        dct['falseValue'] = str(args[0])
        dct['trueValue'] = str(args[1])

        def doStr(self):
            if self:
                return self.trueValue
            else:
                return self.falseValue

        dct['__str__'] = doStr
        if dct['strFmt']:
            print('Bool: ignoring strFmt metadata')

    def new(cls, value):
        """
        Initializes a new boolean instance

        Value must be one of the true/false labels or else a True/False literal
        """
        cls.validate(value)
        # use (value == True) instead of (value is True) so that 0,1
        # can be used for False,True
        if value == True or value == cls.trueValue:  # noqa
            return int.__new__(cls, True)
        elif value == False or value == cls.falseValue:   # noqa
            return int.__new__(cls, False)
        else:
            raise ValueError('Invalid Bool value: %r' % value)

    def addDescriptors(cls):
        cls.descriptors.append(('False', cls.falseValue))
        cls.descriptors.append(('True', cls.trueValue))


# Bitfield value type
class Bits(UInt):

    fieldSpec = re.compile('([a-zA-Z0-9_]+)?(?::([0-9]+))?$')

    @staticmethod
    def binary(value, width):
        return ''.join([str((value >> shift) & 1) for shift in range(width - 1, -1, -1)])

    @classmethod
    def init(cls, dct, *args, **kwargs):
        if not args:
            raise ValueTypeError('missing bitfield specs in ctor')
        offset = 0
        fields = {}
        specs = []
        for field in args:
            parsed = cls.fieldSpec.match(field)
            if not parsed or not parsed.end() == len(field):
                raise ValueTypeError('invalid bitfield spec: %s' % field)
            (name, width) = parsed.groups()
            width = int(width or 1)
            if name:
                if name == 'native':
                    raise ValueTypeError("'native' is not an allowed bitfield name")
                specs.append((name, width))
                fields[name] = (offset, int((1 << width) - 1))
            offset += width
            if offset > 32:
                raise ValueTypeError('total bitfield length > 32')
        dct['width'] = offset
        dct['fieldSpecs'] = specs
        dct['bitFields'] = fields

        def getAttr(self, name):
            if name not in self.bitFields:
                raise AttributeError('no such bitfield "%s"' % name)
            (offset, mask) = self.bitFields[name]
            return (self >> offset) & mask

        dct['__getattr__'] = getAttr

        def setAttr(self, name, value):
            if name not in self.bitFields:
                raise AttributeError('no such bitfield "%s"' % name)
            (offset, mask) = self.bitFields[name]
            return self.__class__((self & ~(mask << offset)) | ((value & mask) << offset))

        dct['set'] = setAttr

        def doRepr(self):
            return '(%s)' % ','.join(
                ['%s=%s' % (n, Bits.binary(getAttr(self, n), w)) for (n, w) in self.fieldSpecs])

        dct['__repr__'] = doRepr

        def bitsString(self):
            return Bits.binary(self, offset)

        dct['bitsString'] = bitsString
        # dct['__str__'] = doStr
        if dct['strFmt']:
            print('Bits: ignoring strFmt metadata')

    def addDescriptors(cls):
        for index, (name, width) in enumerate(cls.fieldSpecs):
            offset, mask = cls.bitFields[name]
            shifted = Bits.binary(mask << offset, cls.width)
            cls.descriptors.append(('Field-%d' % index, '%s %s' % (shifted, name)))


class ByName(object):
    """
    Placeholder for type referred to by name
    """

    def __init__(self, name):
        self.name = name



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2019-05-20
# @Filename: base.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import abc
import asyncio
import inspect
import logging
import pathlib
import time

from sdsstools import get_logger, read_yaml_file

from .tools import REPLY


__all__ = ['BaseClient', 'BaseActor']


class BaseClient(metaclass=abc.ABCMeta):
    """A base client that can be used for listening or for an actor.

    This class defines a new client. Clients differ from actors in that
    they do not receive commands or issue replies, but do send commands to
    other actors and listen to the keyword-value flow. All actors are also
    clients and any actor should subclass from `.BaseClient`.

    Normally a new instance of a client or actor is created by passing a
    configuration file path to `.from_config` which defines how the
    client must be started.

    Parameters
    ----------
    name : str
        The name of the actor.
    version : str
        The version of the actor.
    loop
        The event loop. If `None`, the current event loop will be used.
    log_dir : str
        The directory where to store the file logs.
    log : ~logging.Logger
        A `~logging.Logger` instance to be used for logging instead of creating
        a new one.
    verbose : bool or int
        Whether to log to stdout. Can be an integer logging level.

    """

    name = None

    def __init__(self, name, version=None, loop=None,
                 log_dir=None, log=None, verbose=False):

        self.loop = loop or asyncio.get_event_loop()

        self.name = name
        assert self.name, 'name cannot be empty.'

        self.log = None
        self.setup_logger(log, log_dir, verbose=verbose)

        self.version = version or '?'

        # Internally store the original configuration used to start the client.
        self._config = None

    def __repr__(self):

        return f'<{str(self)} (name={self.name!r})>'

    def __str__(self):

        return self.__class__.__name__

    @abc.abstractmethod
    async def start(self):
        """Runs the client."""

        pass

    async def shutdown(self):
        """Shuts down all the remaining tasks."""

        self.log.info('cancelling all pending tasks and shutting down.')

        tasks = [task for task in asyncio.Task.all_tasks(loop=self.loop)
                 if task is not asyncio.tasks.Task.current_task(loop=self.loop)]
        list(map(lambda task: task.cancel(), tasks))

        await asyncio.gather(*tasks, return_exceptions=True)

        self.loop.stop()

    @staticmethod
    def _parse_config(config):

        if not isinstance(config, dict):

            config = pathlib.Path(config)
            assert config.exists(), 'configuration path does not exist.'

            config = read_yaml_file(str(config))

        if 'actor' in config:
            config = config['actor']

        return config

    @classmethod
    def from_config(cls, config, *args, **kwargs):
        """Parses a configuration file.

        Parameters
        ----------
        config : dict or str
            A configuration dictionary or the path to a YAML configuration
            file that must contain a section ``'actor'`` (if the section is
            not present, the whole file is assumed to be the actor
            configuration).

        """

        orig_config_dict = cls._parse_config(config)
        config_dict = orig_config_dict.copy()

        # Decide what to do with the rest of the keyword arguments:
        args_inspect = inspect.getfullargspec(cls)

        if args_inspect.varkw is not None:
            # If there is a catch-all kw variable, send everything and let the
            # subclass handle it.
            config_dict.update(kwargs)
        else:
            # Check the kw arguments in the subclass and pass only
            # values from config_dict that match them.
            kw_args = args_inspect.kwonlyargs
            if len(args_inspect.defaults) > 0:
                args_invert = args_inspect.args[::-1]
                kw_args += args_invert[:len(args_inspect.defaults)]
            for kw in kwargs:
                if kw in kw_args:
                    config_dict[kw] = kwargs[kw]

        # We also pass *args in case the actor has been subclassed
        # and the subclass' __init__ accepts different arguments.
        new_actor = cls(*args, **config_dict)

        # Store original config. This may not be complete since from_config
        # may have been super'd from somewhere else.
        new_actor._config = orig_config_dict
        new_actor._config.update(kwargs)

        return new_actor

    def setup_logger(self, log, log_dir, verbose=False):
        """Starts the file logger."""

        if not log:
            log = get_logger('actor:' + self.name)

        if log_dir:

            log_dir = pathlib.Path(log_dir).expanduser()

            if not log_dir.exists():
                log_dir.mkdir(parents=True)

            log.start_file_logger(log_dir / f'{self.name}.log')

            log.fh.formatter.converter = time.gmtime
            log.fh.setLevel(REPLY)

        log.sh.setLevel(logging.INFO)
        if verbose:
            log.sh.setLevel(int(verbose))
        else:
            log.sh.setLevel(logging.ERRO)

        self.log = log
        self.log.debug(f'{self.name}: logging system initiated.')

        # Set the loop exception handler to be handled by the logger.
        self.loop.set_exception_handler(self.log.asyncio_exception_handler)

        return log

    def send_command(self):
        """Sends a command to an actor."""

        raise NotImplementedError('Sending commands is not implemented '
                                  'for this client.')


class BaseActor(BaseClient):
    """An actor based on `asyncio`.

    This class expands `.BaseClient` with a parsing system for new commands
    and placeholders for methods for handling new commands and writing replies,
    which should be overridden by the specific actors.

    """

    @abc.abstractmethod
    async def start(self):
        """Starts the server. Must be overridden by the subclasses."""

        pass

    @abc.abstractmethod
    def new_command(self):
        """Handles a new command.

        Must be overridden by the subclass and call `.parse_command`
        with a `.Command` object.

        """

        pass

    @abc.abstractmethod
    def parse_command(self, command):
        """Parses and executes a `.Command`. Must be overridden."""

        pass

    def send_command(self):
        """Sends a command to another actor."""

        raise NotImplementedError('Sending commands is not implemented '
                                  'for this actor.')

    @abc.abstractmethod
    def write(self):
        """Writes a message to user(s). To be overridden by the subclasses."""

        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2018-01-19
# @Filename: command.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import asyncio

from .exceptions import CluError


try:
    import aio_pika as apika
    import aiormq
except ImportError:
    apika = None


__all__ = ['TCPProtocol', 'PeriodicTCPServer',
           'TCPStreamServer', 'TCPStreamPeriodicServer',
           'TCPStreamClient', 'TCPStreamClientContainer',
           'TopicListener']


class TCPProtocol(asyncio.Protocol):
    """A TCP server/client based on asyncio protocols.

    This is a high-level implementation of the client and server asyncio
    protocols. See `asyncio protocol
    <https://docs.python.org/3/library/asyncio-protocol.html>`__ for details.

    Parameters
    ----------
    loop
        The event loop. The current event loop is used by default.
    connection_callback
        Callback to call when a new client connects.
    data_received_callback
        Callback to call when a new data is received.
    max_connections : int
        How many clients the server accepts. If `None`, unlimited connections
        are allowed.

    """

    def __init__(self, loop=None, connection_callback=None,
                 data_received_callback=None, max_connections=None):

        self.connection_callback = connection_callback
        self.data_received_callback = data_received_callback

        self.transports = []
        self.max_connections = max_connections

        self.loop = loop or asyncio.get_event_loop()

    @classmethod
    async def create_server(cls, host, port, **kwargs):
        """Returns a `~asyncio.Server` connection."""

        loop = kwargs.get('loop', asyncio.get_event_loop())

        new_tcp = cls(**kwargs)

        server = await loop.create_server(lambda: new_tcp, host, port)

        await server.start_serving()

        return server

    @classmethod
    async def create_client(cls, host, port, **kwargs):
        """Returns a `~asyncio.Transport` and `~asyncio.Protocol`."""

        if 'connection_callback' in kwargs:
            raise KeyError('connection_callback not allowed when creating a client.')

        loop = kwargs.get('loop', asyncio.get_event_loop())

        new_tcp = cls.__new__(cls, **kwargs)
        transport, protocol = await loop.create_connection(lambda: new_tcp, host, port)

        return transport, protocol

    def connection_made(self, transport):
        """Receives a connection and calls the connection callback."""

        if self.max_connections is None or (len(self.transports) < self.max_connections):
            self.transports.append(transport)
        else:
            transport.write('Maximum number of connections reached.')
            transport.close()

        if self.connection_callback:
            self.connection_callback(transport)

    def data_received(self, data):
        """Decodes the received data."""

        if self.data_received_callback:
            self.data_received_callback(data.decode())

    def connection_lost(self, exc):
        """Called when connection is lost."""

        pass


class PeriodicTCPServer(TCPProtocol):
    """A TCP server that runs a callback periodically.

    Parameters
    ----------
    period_callback
        Callback to run every iteration.
    sleep_time : float
        The delay between two calls to ``periodic_callback``.
    kwargs : dict
        Parameters to pass to `TCPProtocol`

    """

    def __init__(self, periodic_callback=None, sleep_time=1, **kwargs):

        self._periodic_callback = periodic_callback
        self.sleep_time = sleep_time

        self.periodic_task = None

        super().__init__(**kwargs)

    @classmethod
    async def create_client(cls, *args, **kwargs):

        raise NotImplementedError('create_client is not implemented for PeriodicTCPServer.')

    @classmethod
    async def create_server(cls, host, port, *args, **kwargs):
        """Returns a `~asyncio.Server` connection."""

        loop = kwargs.get('loop', asyncio.get_event_loop())

        new_tcp = cls(*args, **kwargs)

        server = await loop.create_server(lambda: new_tcp, host, port)

        await server.start_serving()

        new_tcp.periodic_task = asyncio.create_task(new_tcp._emit_periodic())

        return server

    @property
    def periodic_callback(self):
        """Returns the periodic callback."""

        return self._periodic_callback

    @periodic_callback.setter
    def periodic_callback(self, func):
        """Sets the periodic callback."""

        self._periodic_callback = func

    async def _emit_periodic(self):

        while True:

            if self.periodic_callback is not None:
                for transport in self.transports:
                    if asyncio.iscoroutinefunction(self.periodic_callback):
                        await self.periodic_callback(transport)
                    else:
                        self.periodic_callback(transport)

            await asyncio.sleep(self.sleep_time)


class TCPStreamServer(object):
    """A TCP server based on asyncio streams.

    This is a high-level implementation of the asyncio server using
    streams. See `asyncio streams
    <https://docs.python.org/3/library/asyncio-stream.html>`__ for details.

    Parameters
    ----------
    host : str
        The server host.
    port : int
        The server port.
    connection_callback
        Callback to call when a new client connects or disconnects.
    data_received_callback
        Callback to call when a new data is received.
    loop
        The event loop. The current event loop is used by default.
    max_connections : int
        How many clients the server accepts. If `None`, unlimited connections
        are allowed.

    """

    def __init__(self, host, port, connection_callback=None,
                 data_received_callback=None, loop=None, max_connections=None):

        self.host = host
        self.port = port

        self.transports = {}
        self.loop = loop or asyncio.get_event_loop()

        self.max_connections = max_connections

        self.connection_callback = connection_callback
        self.data_received_callback = data_received_callback

        #: The `asyncio.Server`. Created when `.start_server` is run.
        self.server = None

    async def start_server(self):
        """Starts the server and returns a `~asyncio.Server` connection."""

        self.server = await asyncio.start_server(self.connection_made,
                                                 self.host, self.port)

        return self.server

    def serve_forever(self):
        """Exposes ``TCPStreamServer.server.serve_forever``."""

        return self.server.serve_forever()

    async def _do_callback(self, cb, *args, **kwargs):
        """Calls a function or coroutine callback."""

        if asyncio.iscoroutinefunction(cb):
            return await asyncio.create_task(cb(*args, **kwargs))
        else:
            return cb(*args, **kwargs)

    async def connection_made(self, reader, writer):
        """Called when a new client connects to the server.

        Stores the writer protocol in ``transports``, calls the connection
        callback, if any, and starts a loop to read any incoming data.

        """

        if self.max_connections and len(self.transports) == self.max_connections:
            writer.write('Max number of connections reached.\n'.encode())
            return

        self.transports[writer.transport] = writer

        if self.connection_callback:
            await self._do_callback(self.connection_callback, writer.transport)

        while True:

            try:
                data = await reader.readuntil()
            except asyncio.IncompleteReadError:
                writer.close()
                self.transports.pop(writer.transport)
                if self.connection_callback:
                    await self._do_callback(self.connection_callback, writer.transport)
                break

            if self.data_received_callback:
                await self._do_callback(self.data_received_callback,
                                        writer.transport, data)


class TCPStreamClientContainer(object):
    """An object containing a writer and reader stream to a TCP server."""

    def __init__(self, host, port):

        self.host = host
        self.port = port

        self.reader = None
        self.writer = None

    async def open_connection(self):
        """Creates the connection."""

        self.reader, self.writer = await asyncio.open_connection(self.host, self.port)

    def close(self):
        """Closes the stream."""

        if self.writer:
            self.writer.close()
        else:
            raise RuntimeError('connection cannot be closed because it is not open.')


async def TCPStreamClient(host, port):
    """Returns a TCP stream connection with a writer and reader.

    This function is more convenient than doing ::

        >>> client = TCPStreamClientContainer('127.0.0.1', 5555)
        >>> await client.open_connection()

    Instead just do ::

        >>> client = await TCPStreamClient('127.0.0.1', 5555)
        >>> client.writer.write('Hi!\\n'.encode())

    Parameters
    ----------
    host : str
        The host of the TCP server.
    port : int
        The port of the TCP server.

    Returns
    -------
    client : `.TCPStreamClientContainer`
        A container for the stream reader and writer.

    """

    client = TCPStreamClientContainer(host, port)
    await client.open_connection()

    return client


class TCPStreamPeriodicServer(TCPStreamServer):
    """A TCP server that calls a function periodically.

    Parameters
    ----------
    host : str
        The server host.
    port : int
        The server port.
    period_callback
        Callback to run every iteration.
    sleep_time : float
        The delay between two calls to ``periodic_callback``.
    kwargs : dict
        Parameters to pass to `TCPStreamServer`

    """

    def __init__(self, host, port, periodic_callback=None,
                 sleep_time=1, **kwargs):

        self._periodic_callback = periodic_callback
        self.sleep_time = sleep_time

        self.periodic_task = None

        super().__init__(host, port, **kwargs)

    async def start_server(self):
        """Starts the server and returns a `~asyncio.Server` connection."""

        server = await super().start_server()

        self.periodic_task = asyncio.create_task(self._emit_periodic())

        return server

    @property
    def periodic_callback(self):
        """Returns the periodic callback."""

        return self._periodic_callback

    @periodic_callback.setter
    def periodic_callback(self, func):
        """Sets the periodic callback."""

        self._periodic_callback = func

    async def _emit_periodic(self):

        while True:

            if self.server and self.periodic_callback:
                for transport in self.transports:
                    await self._do_callback(self.periodic_callback, transport)

            await asyncio.sleep(self.sleep_time)


class TopicListener(object):
    """A class to declare and listen to AMQP queues with topic conditions.

    Parameters
    ----------
    user : str
        The user to connect to the RabbitMQ broker.
    host : str
        The host where the RabbitMQ message broker lives.

    """

    def __init__(self, user=None, host=None):

        if not apika:
            raise ImportError('cannot use TopicListener without aoi_pika.')

        self.user = user
        self.host = host

        self.connection = None
        self.channel = None
        self.exchange = None
        self.queues = []

    async def connect(self, exchange_name, channel=None, loop=None,
                      exchange_type=apika.ExchangeType.TOPIC):
        """Initialise the connection.

        Parameters
        ----------
        exchange_name : str
            The name of the exchange to create.
        channel
            If specified, ``user`` and ``host`` are ignored and the connection
            and channel are set from ``channel``.
        exchange_type : str
            The type of exchange to create.
        loop
            Event loop. If empty, the current event loop will be used.

        """

        self.loop = loop or asyncio.get_event_loop()

        if not channel:

            assert self.user and self.host, 'either user or host are not set.'

            self.connection = await apika.connect_robust(user=self.user,
                                                         host=self.host,
                                                         loop=self.loop)

            self.channel = await self.connection.channel()
            await self.channel.set_qos(prefetch_count=1)

        else:

            self.connection = channel._connection
            self.channel = channel

            self.user = self.connection.url.user
            self.host = self.connection.url.host

        self.exchange = await self.channel.declare_exchange(exchange_name,
                                                            type=exchange_type,
                                                            auto_delete=True)

        return self

    async def add_queue(self, queue_name, callback=None, bindings='*'):
        """Adds a queue with bindings.

        Parameters
        ----------
        queue_name : str
            The name of the queue to create.
        callback
            A callable that will be called when a new message is received in
            the queue. Can be a coroutine.
        bindings : list or str
            The list of bindings for the queue. Can be a list of string or a
            single string in which the bindings are comma-separated.

        """

        if isinstance(bindings, str):
            bindings = bindings.split(',')
        elif isinstance(bindings, (list, tuple)):
            bindings = list(bindings)
        else:
            raise TypeError('invalid type for bindings {bindings!r}.')

        try:
            queue = await self.channel.declare_queue(queue_name, exclusive=True)
        except aiormq.exceptions.ChannelLockedResource:
            raise CluError(f'cannot create queue {queue_name}. '
                           'This may indicate that another instance of the '
                           'same actor is running.')

        for binding in bindings:
            await queue.bind(self.exchange, routing_key=binding)

        self.queues.append(queue)

        if callback:
            queue.consumer_task = self.loop.create_task(queue.consume(callback))

        return queue



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2018-09-07
# @Filename: tools.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import asyncio
import collections
import contextlib
import enum
import functools
import json
import logging
import re


REPLY = 5  # REPLY logging level
WARNING_REGEX = r'^.*?\s*?(\w*?Warning): (.*)'


__ALL__ = ['CommandStatus', 'StatusMixIn', 'format_value', 'CallbackScheduler',
           'CaseInsensitiveDict', 'cli_coro', 'value', 'as_complete_failer',
           'log_reply', 'ActorHandler']


class Maskbit(enum.Flag):
    """A maskbit enumeration. Intended for subclassing."""

    @property
    def active_bits(self):
        """Returns a list of flags that match the value."""

        return [bit for bit in self.__class__ if bit.value & self.value]


COMMAND_STATUS_TO_CODE = {
    'DONE': ':',
    'CANCELLED': 'f',
    'FAILED': 'f',
    'TIMEDOUT': 'f',
    'READY': 'i',
    'RUNNING': 'i',
    'CANCELLING': 'w',
    'FAILING': 'w',
    'DEBUG': 'd',
}


class CommandStatus(Maskbit):

    DONE = enum.auto()
    CANCELLED = enum.auto()
    FAILED = enum.auto()
    TIMEDOUT = enum.auto()
    READY = enum.auto()
    RUNNING = enum.auto()
    CANCELLING = enum.auto()
    FAILING = enum.auto()
    DEBUG = enum.auto()

    ACTIVE_STATES = RUNNING | CANCELLING | FAILING
    FAILED_STATES = CANCELLED | FAILED | TIMEDOUT
    FAILING_STATES = CANCELLING | FAILING
    DONE_STATES = DONE | FAILED_STATES
    ALL_STATES = READY | ACTIVE_STATES | DONE_STATES

    def __init__(self, *args):

        if self.name.upper() in COMMAND_STATUS_TO_CODE:
            self.code = COMMAND_STATUS_TO_CODE[self.name.upper()]
        else:
            self.code = None

    @property
    def is_combination(self):
        """Returns True if a flag is a combination."""

        if bin(self).count('1') > 1:
            return True
        return False

    @property
    def did_fail(self):
        """Command failed or was cancelled."""

        return self in self.FAILED_STATES

    @property
    def did_succeed(self):
        """Command finished with DONE status."""

        return self == self.DONE

    @property
    def is_active(self):
        """Command is running, cancelling or failing."""

        return self in self.ACTIVE_STATES

    @property
    def is_done(self):
        """Command is done (whether successfully or not)."""

        return self in self.DONE_STATES

    @property
    def is_failing(self):
        """Command is being cancelled or is failing."""

        return self in self.FAILING_STATES

    @staticmethod
    def get_inverse_dict():
        """Gets a reversed dictionary of code to status.

        Note that the inverse dictionary is not unique and you can get
        different statuses associated with the same code.

        """

        return dict((status.code, status) for status in CommandStatus if status.code)


class StatusMixIn(object):
    """A mixin that provides status tracking with callbacks.

    Provides a status property that executes a list of callbacks when
    the status changes.

    Parameters
    ----------
    maskbit_flags : class
        A class containing the available statuses as a series of maskbit
        flags. Usually as subclass of `enum.Flag`.
    initial_status : str
        The initial status.
    callback_func : function
        The function to call if the status changes.
    call_now : bool
        Whether the callback function should be called when initialising.

    Attributes
    ----------
    callbacks : list
        A list of the callback functions to call.

    """

    def __init__(self, maskbit_flags, initial_status=None,
                 callback_func=None, call_now=False):

        self.flags = maskbit_flags
        self.callbacks = []
        self._status = initial_status
        self.watcher = None

        if callback_func is not None:
            if isinstance(callback_func, (list, tuple)):
                self.callbacks = callback_func
            else:
                self.callbacks.append(callback_func)

        if call_now is True:
            self.do_callbacks()

    def do_callbacks(self):
        """Calls functions in ``callbacks``."""

        assert hasattr(self, 'callbacks'), 'missing callbacks attribute.'

        loop = self.loop if hasattr(self, 'loop') else asyncio.get_event_loop()

        for func in self.callbacks:
            loop.call_soon(func)

    @property
    def status(self):
        """Returns the status."""

        return self._status

    @status.setter
    def status(self, value):
        """Sets the status."""

        if value != self._status:
            self._status = self.flags(value)
            self.do_callbacks()
            if self.watcher is not None:
                self.watcher.set()

    async def wait_for_status(self, value, loop=None):
        """Awaits until the status matches ``value``."""

        if self.status == value:
            return

        if loop is None:
            if hasattr(self, 'loop') and self.loop is not None:
                loop = self.loop
            else:
                loop = asyncio.get_event_loop()

        self.watcher = asyncio.Event(loop=loop)

        while self.status != value:
            await self.watcher.wait()
            if self.watcher is not None:
                self.watcher.clear()

        self.watcher = None


class CallbackScheduler(object):
    """A queue for executing callbacks."""

    def __init__(self, loop=None):

        self.loop = loop or asyncio.get_event_loop()
        self.queue = asyncio.Queue()

        self.running = []  # Running callbacks
        self._task = self.loop.create_task(self._process_queue())

    async def stop(self):
        """Stops processing callbacks and awaits currently running ones."""

        self._task.cancel()

        for cb in self.running:
            if not cb.done():
                cb.cancel()

        with contextlib.suppress(asyncio.CancelledError):
            await self._task
            for cb in self.running:
                await cb

        self.running = []

    def add_callback(self, cb, *args, **kwargs):
        """Add a callback to the queue.

        The callback will be called as ``cb(*args, **kwargs).

        """

        self.queue.put_nowait((cb, args, kwargs))

    async def _process_queue(self):
        """Processes new callbacks."""

        while True:

            cb, args, kwargs = await self.queue.get()

            if asyncio.iscoroutinefunction(cb):
                self.running.append(asyncio.create_task(cb(*args, **kwargs)))
            else:
                self.loop.call_soon(functools.partial(cb, *args, **kwargs))

            # Clean already done callbacks
            done_cb = []
            for task in self.running:
                if task.done():
                    done_cb.append(task)

            self.running = [task for task in self.running if task not in done_cb]


def format_value(value):
    """Formats messages in a way that is compatible with the parser.

    Parameters
    ----------
    value
        The data to be formatted.

    Returns
    -------
    formatted_text : `str`
        A string with the escaped text.

    """

    if isinstance(value, str):
        if ' ' in value and not (value.startswith('\'') or value.startswith('"')):
            value = json.dumps(value)
    elif isinstance(value, bool):
        value = 'T' if value else 'F'
    elif isinstance(value, (tuple, list)):
        value = ','.join([format_value(item) for item in value])
    else:
        value = str(value)

    return value


def escape(value):
    """Escapes a text using `json.dumps`."""

    return json.dumps(value)


class CaseInsensitiveDict(collections.OrderedDict):
    """A dictionary that performs case-insensitive operations."""

    def __init__(self, values):

        self._lc = []

        collections.OrderedDict.__init__(self, values)

        self._lc = [key.lower() for key in values]
        assert len(set(self._lc)) == len(self._lc), 'the are duplicated items in the dict.'

    def __get_key__(self, key):
        """Returns the correct value of the key, regardless of its case."""

        try:
            idx = self._lc.index(key.lower())
        except ValueError:
            return key

        return list(self)[idx]

    def __getitem__(self, key):
        return collections.OrderedDict.__getitem__(self, self.__get_key__(key))

    def __setitem__(self, key, value):

        if key.lower() not in self._lc:
            self._lc.append(key.lower())
            collections.OrderedDict.__setitem__(self, key, value)
        else:
            collections.OrderedDict.__setitem__(self, self.__get_key__(key), value)

    def __contains__(self, key):
        return collections.OrderedDict.__contains__(self, self.__get_key__(key))

    def __eq__(self, key):
        return collections.OrderedDict.__eq__(self, self.__get_key__(key))


def cli_coro(f):
    """Decorator function that allows defining coroutines with click."""

    f = asyncio.coroutine(f)

    def wrapper(*args, **kwargs):
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(f(*args, **kwargs))

    return functools.update_wrapper(wrapper, f)


async def as_complete_failer(aws, on_fail_callback=None, **kwargs):
    """Similar to `~asyncio.as_complete` but cancels all the tasks
    if any of them returns `False`.

    Parameters
    ----------
    aws : list
        A list of awaitable objects. If not a list, it will be wrapped in one.
    on_fail_callback
        A function or coroutine to call if any of the tasks failed.
    kwargs : dict
        A dictionary of keywords to be passed to `~asyncio.as_complete`.

    Returns
    -------
    result_tuple : tuple
        A tuple in which the first element is `True` if all the tasks
        completed, `False` if any of them failed and the rest were cancelled.
        If `False`, the second element is `None` if no exceptions were caught
        during the execution of the tasks, otherwise it contains the error
        message. If `True`, the second element is always `None`.

    """

    if not isinstance(aws, (list, tuple)):
        aws = [aws]

    loop = kwargs.get('loop', asyncio.get_event_loop())

    tasks = [loop.create_task(aw) for aw in aws]

    failed = False
    error_message = None
    for next_completed in asyncio.as_completed(tasks, **kwargs):
        try:
            result = await next_completed
        except Exception as ee:
            error_message = str(ee)
            result = False

        if not result:
            failed = True
            break

    if failed:

        # Cancel tasks
        [task.cancel() for task in tasks]

        with contextlib.suppress(asyncio.CancelledError):
            await asyncio.gather(*[task for task in tasks])

        if on_fail_callback:
            if asyncio.iscoroutinefunction(on_fail_callback):
                await on_fail_callback()
            else:
                on_fail_callback()

        return (False, error_message)

    return (True, None)


def log_reply(log, message_code, message, use_message_code=False):
    """Logs an actor message with the correct code."""

    code_dict = {'f': logging.ERROR,
                 'e': logging.ERROR,
                 'w': logging.WARNING,
                 'i': logging.INFO,
                 ':': logging.INFO,
                 'd': logging.DEBUG}

    if use_message_code:
        log.log(code_dict[message_code], message)
    else:
        # Sets the REPLY log level
        log_level_no = REPLY
        if log_level_no in logging._levelToName:
            log_level = log_level_no
        else:
            log_level = logging.DEBUG

        log.log(log_level, message)


class ActorHandler(logging.Handler):
    """A handler that outputs log messages as actor keywords.

    Parameters
    ----------
    actor
        The actor instance.
    level : int
        The level above which records will be output in the actor.
    keyword : str
        The keyword around which the messages will be output.
    code_mapping : dict
        A mapping of logging levels to actor codes. The values provided
        override the default mapping. For example, to make input log messages
        with info level be output as debug,
        ``code_mapping={logging.INFO: 'd'}``.
    filter_warnings : list
        A list of warning classes that will be issued to the actor. Subclasses
        of the filter warning are accepted, any other warnings will be ignored.

    """

    def __init__(self, actor, level=logging.ERROR, keyword='text',
                 code_mapping=None, filter_warnings=None):

        self.actor = actor
        self.keyword = keyword

        self.code_mapping = {logging.DEBUG: 'd',
                             logging.INFO: 'i',
                             logging.WARNING: 'w',
                             logging.ERROR: 'f'}

        if code_mapping:
            self.code_mapping.update(code_mapping)

        self.filter_warnings = filter_warnings

        super().__init__(level=level)

    def emit(self, record):
        """Emits the record."""

        message = record.getMessage()
        message_lines = message.splitlines()

        if record.exc_info:
            message_lines.append(f'{record.exc_info[0].__name__}: {record.exc_info[1]}')

        if record.levelno <= logging.DEBUG:
            code = self.code_mapping[logging.DEBUG]
        elif record.levelno <= logging.INFO:
            code = self.code_mapping[logging.INFO]
        elif record.levelno <= logging.WARNING:
            code = self.code_mapping[logging.WARNING]
            warning_category_groups = re.match(WARNING_REGEX, message)
            if warning_category_groups is not None:
                message_lines = self._filter_warning(warning_category_groups)
        elif record.levelno >= logging.ERROR:
            code = self.code_mapping[logging.ERROR]
        else:
            code = 'w'

        for line in message_lines:
            result = self.actor.write(code, message={self.keyword: line})

            if asyncio.iscoroutine(result):
                asyncio.create_task(result)

    def _filter_warning(self, warning_category_groups):

        warning_category, warning_text = warning_category_groups.groups()
        message_lines = [f'{warning_text} ({warning_category})']

        try:
            warning_class = eval(warning_category)
            if self.filter_warnings:
                for warning_filter in self.filter_warnings:
                    if isinstance(warning_class, warning_filter):
                        return message_lines
            return []
        except NameError:
            return message_lines



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2019-04-22
# @Filename: __init__.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

# flake8: noqa

import logging

from sdsstools import get_package_version

from .actor import *
from .base import *
from .client import *
from .command import *
from .device import *
from .exceptions import *
from .legacy import LegacyActor
from .parser import ClickParser, command_parser
from .tools import (REPLY, ActorHandler, CommandStatus,
                    as_complete_failer, escape, format_value)


# Add REPLY level to logging
logging.addLevelName(REPLY, 'REPLY')
logging.Logger.REPLY = lambda self, message, *args, **kws: \
    self._log(REPLY, message, *args, **kws)


NAME = 'sdss-clu'
__version__ = get_package_version(__file__, 'sdss-clu', pep_440=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-07-30
# @Filename: client.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import json
import uuid

from .base import BaseClient
from .command import Command
from .model import Reply
from .protocol import TopicListener
from .tools import CommandStatus


try:
    import aio_pika as apika
except ImportError:
    apika = None


__all__ = ['AMQPClient']


class AMQPClient(BaseClient):
    """Defines a new client based on the AMQP standard.

    To start a new client first instantiate the class and then run `.run` as
    a coroutine. Note that `.run` does not block so you will need to use
    asyncio's ``run_forever`` or a similar system ::

        >>> loop = asyncio.get_event_loop()
        >>> client = await Client('my_client', 'guest', 'localhost', loop=loop).start()
        >>> loop.run_forever()

    Parameters
    ----------
    name : str
        The name of the actor.
    user : str
        The user to connect to the AMQP broker.
    host : str
        The host where the AMQP message broker lives.
    version : str
        The version of the actor.
    loop
        The event loop. If `None`, the current event loop will be used.
    log_dir : str
        The directory where to store the logs. Defaults to
        ``$HOME/logs/<name>`` where ``<name>`` is the name of the actor.
    log : ~logging.Logger
        A `~logging.Logger` instance to be used for logging instead of creating
        a new one.
    parser : ~clu.parser.CluGroup
        A click command parser that is a subclass of `~clu.parser.CluGroup`.
        If `None`, the active parser will be used.

    """

    __EXCHANGE_NAME__ = 'actor_exchange'

    connection = None

    def __init__(self, name, user, host, version=None,
                 loop=None, log_dir=None, log=None, parser=None):

        if not apika:
            raise ImportError('to instantiate a new Client class '
                              'you need to have aio_pika installed.')

        super().__init__(name, version=version, loop=loop,
                         log_dir=log_dir, log=log, parser=parser)

        self.user = user
        self.host = host

        self.replies_queue = None

        # Creates the connection to the AMQP broker
        self.connection = TopicListener(self.user, self.host)

        #: dict: External commands currently running.
        self.running_commands = {}

    def __repr__(self):

        if not self.connection or self.connection.connection is None:
            url = 'disconnected'
        else:
            url = str(self.connection.connection.url)

        return f'<{str(self)} (name={self.name!r}, {url}>'

    async def start(self, exchange_name=__EXCHANGE_NAME__):
        """Starts the connection to the AMQP broker."""

        # Starts the connection and creates the exchange
        await self.connection.connect(exchange_name, loop=self.loop)

        # Binds the replies queue.
        self.replies_queue = await self.connection.add_queue(
            f'{self.name}_replies', callback=self.handle_reply,
            bindings=['reply.#'])

        self.log.info(f'replies queue {self.replies_queue.name!r} '
                      f'bound to {self.connection.connection.url!s}')

        return self

    @classmethod
    def from_config(cls, config, *args, **kwargs):
        """Starts a new client from a configuration file.

        Refer to `.BaseClient.from_config`.

        """

        config_dict = cls._parse_config(config)

        args = list(args) + [config_dict.pop('name'),
                             config_dict.pop('user'),
                             config_dict.pop('host')]

        return super().from_config(config_dict, *args, **kwargs)

    async def handle_reply(self, message):
        """Handles a reply received from the exchange.

        Creates a new instance of `.Reply` from the ``message``. If the
        reply is valid it updates any running command.

        Parameters
        ----------
        message : aio_pika.IncomingMessage
            The message received.

        Returns
        -------
        reply : `.Reply`
            The `.Reply` object created from the message.

        """

        reply = Reply(message, ack=True)
        if not reply.is_valid:
            self.log.error('invalid message.')
            return reply

        # Ignores message from self.
        if reply.sender and self.name == reply.sender:
            return reply

        # If the command is running we check if the message code indicates
        # the command is done and, if so, sets the result in the Future.
        if reply.command_id in self.running_commands:
            is_done = CommandStatus.get_inverse_dict()[reply.message_code].is_done
            if is_done:
                command = self.running_commands.pop(reply.command_id)
                if not command.done():
                    command.set_result(command)

        return reply

    async def send_command(self, consumer, command_string, command_id=None):
        """Commands another actor over its RCP queue.

        Parameters
        ----------
        consumer : str
            The actor we are commanding.
        command_string : str
            The command string that will be parsed by the remote actor.
        command_id
            The command ID associated with this command. If empty, an unique
            identifier will be attached.

        """

        command_id = command_id or str(uuid.uuid4())

        # Creates and registers a command.
        command = Command(command_string=command_string,
                          command_id=command_id,
                          commander_id=self.name,
                          consumer_id=consumer,
                          actor=self, loop=self.loop)

        self.running_commands[command_id] = command

        headers = {'command_id': command_id,
                   'commander_id': self.name}

        # The routing key has the topic command and the name of the commanded actor.
        routing_key = f'command.{consumer}'

        message_body = {'command_string': command_string}

        await self.connection.exchange.publish(
            apika.Message(json.dumps(message_body).encode(),
                          content_type='text/json',
                          headers=headers,
                          correlation_id=command_id,
                          reply_to=self.replies_queue.name),
            routing_key=routing_key)

        return command



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2019-05-17
# @Filename: model.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import json
import pathlib

import jsonschema

from .tools import CallbackScheduler, CaseInsensitiveDict


__all__ = ['Reply', 'Property', 'BaseModel', 'Model', 'ModelSet']


class Reply(object):
    """A container for a `~aio_pika.IncomingMessage` that expands and decodes it.

    Parameters
    ----------
    message : aio_pika.IncomingMessage
        The message that contains the reply.
    ack : bool
        Whether to acknowledge the message.

    Attributes
    ----------
    is_valid : bool
        Whether the message is valid and correctly parsed.
    body : dict
        The body of the message, as a JSON dictionary.
    info : dict
        The info dictionary.
    headers : dict
        The headers of the message, decoded if they are bytes.
    message_code : str
        The message code.
    sender : str
        The name of the actor that sends the reply.
    command_id
        The command ID.

    """

    def __init__(self, message, ack=True):

        self.message = message

        self.is_valid = True

        self.body = None

        # Acknowledges receipt of message
        if ack:
            message.ack()

        self.info = message.info()

        self.headers = self.info['headers']
        for key in self.headers:
            if isinstance(self.headers[key], bytes):
                self.headers[key] = self.headers[key].decode()

        self.message_code = self.headers.get('message_code', None)
        if self.message_code is None:
            self.log.warning(f'received message without message_code: {message}')

        self.sender = self.headers.get('sender', None)
        if self.sender is None:
            self.log.warning(f'received message without sender: {message}')

        self.command_id = message.correlation_id

        command_id_header = self.headers.get('command_id', None)
        if command_id_header and command_id_header != self.command_id:
            self.log.error(f'mismatch between message '
                           f'correlation_id={self.command_id} '
                           f'and header command_id={command_id_header} '
                           f'in message {message}')
            self.is_valid = False
            return

        self.body = json.loads(self.message.body.decode())


class Property(object):
    """A model property with callbacks.

    Parameters
    ----------
    key
        The property to be represented.
    model : BaseModel
        The parent model.
    callback
        The function or coroutine that will be called if the value of the key
        if modified. The callback is called with the instance of `BaseProperty`
        as the only argument. Note that the callback will be scheduled even
        if the new value is the same as the previous one.

    """

    def __init__(self, key, model=None, callback=None):

        self.key = key
        self._value = None

        self.model = model

        self.scheduler = CallbackScheduler()
        self.callback = callback

    def __repr__(self):
        return f'<{self.__class__.__name__!s} ({self.key}): {self.value}>'

    def __str__(self):
        return str(self.value)

    @property
    def value(self):
        """The value associated to the key."""

        return self._value

    @value.setter
    def value(self, new_value):
        """Sets the value of the key and schedules the callback."""

        self._value = new_value

        if self.callback:
            self.scheduler.add_callback(self.callback, self)


class BaseModel(CaseInsensitiveDict):
    """A JSON-compliant model.

    Parameters
    ----------
    name : str
        The name of the model.
    callback
        A function or coroutine to call when the datamodel changes. The
        function is called with the instance of `BaseModel` as the only
        parameter. If the callback is a coroutine, it is scheduled as a task.
    log : ~logging.Logger
        Where to log messages.

    """

    def __init__(self, name, callback=None, log=None):

        self.name = name

        self.callback = callback
        self.scheduler = CallbackScheduler()

        self.log = log

        CaseInsensitiveDict.__init__(self, {})

    def __repr__(self):
        return f'<Model ({self.name})>'

    def __str__(self):
        return str(self.flatten())

    def flatten(self):
        """Returns a dictionary of values.

        Return a dictionary in which the `Property` instances are replaced
        with their values.

        """

        return {key: self[key].value for key in self}

    def jsonify(self):
        """Returns a JSON string with the model."""

        return json.dumps(self.unkey())


class Model(BaseModel):
    """A model with JSON validation.

    In addition to the parameters in `.BaseModel`, the following parameters
    are accepted:

    Parameters
    ----------
    schema : dict
        A valid JSON schema, to be used for validation.

    """

    VALIDATOR = jsonschema.Draft7Validator

    def __init__(self, name, schema, **kwargs):

        self.schema = schema

        if not self.check_schema(schema, is_file=False):
            raise ValueError(f'schema {name!r} is invalid.')

        self.validator = self.VALIDATOR(self.schema)

        super().__init__(name, **kwargs)

        for key in self.schema['properties']:
            self[key] = Property(key, model=self)

    @staticmethod
    def check_schema(schema, is_file=False):
        """Checks whether a JSON schema is valid.

        Parameters
        ----------
        schema : str or dict
            The schema to check. It can be a JSON dictionary or the path to a
            file.
        is_file : bool
            Whether the input schema is a filepath or not.

        Returns
        -------
        result : `bool`
            Returns `True` if the schema is a valid JSON schema, `False`
            otherwise.

        """

        if is_file:
            schema = json.load(open(pathlib.Path(schema).expanduser(), 'r'))
        elif not is_file and isinstance(schema, str):
            try:
                schema = json.loads(schema)
            except json.JSONDecodeError:
                raise ValueError('cannot parse input schema.')

        try:
            Model.VALIDATOR.check_schema(schema)
            return True
        except jsonschema.SchemaError:
            return False

    def update_model(self, instance):
        """Validates a new instance and updates the model."""

        try:
            self.validator.validate(instance)
        except jsonschema.exceptions.ValidationError:
            if self.log:
                self.log.error(f'model {self.name} cannot be updated. '
                               f'Failed validating instance {instance}.')
            return False

        for key, value in instance.items():
            self[key].value = value

        if self.callback:
            self.scheduler.add_callback(self.callback, self)

        return True


class ModelSet(dict):
    """A dictionary of `Model` instances from files.

    Reads model schemas from files and creates a dictionary of `Model`
    instances.

    Parameters
    ----------
    model_path : str or pathlib.Path
        The path to the directory containing the schema files. Each schema
        file must be named as the model and have extension ``.json``
        (e.g., ``sop.json``).
    model_names : list
        A list of models whose schemas will be loaded.
    raise_exception : bool
        Whether to raise an exception if any of the models cannot be loaded.
    kwargs
        Keyword arguments to be passed to `Model`.

    Example
    -------

        >>> model_set = ModelSet('~/my_models', model_names=['sop', 'guider'])
        >>> model_set['sop']
        <Model (name='sop')>

    """

    def __init__(self, model_path, model_names, raise_exception=True, **kwargs):

        dict.__init__(self, {})

        self.log = kwargs.get('log', None)

        self.model_path = pathlib.Path(model_path).expanduser()

        for name in model_names:

            try:

                schema_path = self.model_path / f'{name}.json'
                assert schema_path.exists(), f'model path {schema_path} does not exist.'

                schema = json.load(open(schema_path))

                self[name] = Model(name, schema, **kwargs)

            except Exception as ee:

                if not raise_exception:
                    if self.log:
                        self.log.warning(f'cannot load model for actor {name!r}: {ee}')
                    continue
                raise



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2018-01-17
# @Filename: command.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import asyncio
import re

import clu
from clu.tools import CommandStatus, StatusMixIn


__all__ = ['BaseCommand', 'Command', 'parse_legacy_command']


class BaseCommand(asyncio.Future, StatusMixIn):
    """Base class for commands of all types (user and device).

    A `BaseCommand` instance is a `~asyncio.Future` whose result gets set
    when the status is done (either successfully or not).

    Parameters
    ----------
    cmd_str : str
        The command string to be parsed.
    commander_id : str or int
        The ID of the commander issuing this command. Can be a string or an
        integer. Normally the former is used for new-style actor and the latter
        for legacy actors.
    command_id : str or int
        The ID associated to this command. As with the commander_id, it can be
        a string or an integer
    consumer_id : str or int
        The actor that is consuming this command. Normally this is our own
        actor but if we are commanding another actor ``consumer_id`` will
        be the destination actor.
    parent : .BaseCommand
        Another `.BaseCommand` object that is issuing this subcommand. If they
        are not specified, sets ``commander_id`` and ``consumer_id``
        appropriately.
    status_callback : function
        A function to call when the status changes.
    call_now : bool
        Whether to call ``status_callback`` when initialising the command.
    default_keyword : str
        The keyword to use when writing a message that does not specify a
        keyword.
    loop
        The event loop.

    """

    def __init__(self, commander_id=0, command_id=0, consumer_id=0, parent=None,
                 status_callback=None, call_now=False, default_keyword='text',
                 loop=None):

        self.commander_id = commander_id
        self.consumer_id = consumer_id
        self.command_id = command_id

        self.parent = parent

        if self.parent:
            if not isinstance(self.parent, BaseCommand):
                raise clu.CluError('invalid parent type.')
            if self.parent.command_id != 0 and self.command_id == self.parent.command_id:
                raise clu.CluError('subcommand cannot have the '
                                   'same command_id as the parent.')

            self.commander_id = self.commander_id or self.parent.commander_id
            self.consumer_id = self.consumer_id or self.parent.consumer_id

        self.default_keyword = default_keyword
        self.loop = loop or asyncio.get_event_loop()

        asyncio.Future.__init__(self, loop=self.loop)

        StatusMixIn.__init__(self, CommandStatus, initial_status=CommandStatus.READY,
                             callback_func=status_callback, call_now=call_now)

    @property
    def status(self):
        """Returns the status."""

        return self._status

    @status.setter
    def status(self, status):
        """Sets the status. A message is output to the users.

        This setter calls `.set_status` with an empty message to the users.

        Parameters
        ----------
        status : CommandStatus or int or str
            The status to set, either as a `CommandStatus` value or the
            integer associated with the maskbit. If ``value`` is a string,
            loops over the bits in `CommandStatus` and assigns the one whose
            name matches.

        """

        self.set_status(status)

    def set_status(self, status, message=None, **kwargs):
        """Same as `.status` but allows to specify a message to the users."""

        if self.status.is_done:
            raise RuntimeError('cannot modify a done command.')

        if not isinstance(status, self.flags):
            if isinstance(status, int):
                status = self.flags(int)
            elif isinstance(status, str):
                for bit in self.flags:
                    if status.lower() == bit.name.lower():
                        status = bit
                        break
            else:
                raise ValueError(f'status {status!r} is not a valid command status.')

        if status != self._status:

            status_code = status.code

            self.write(status_code, message, **kwargs)

            self._status = status

            self.do_callbacks()

            # If the command is done, set the result of the future.
            if self._status.is_done:
                self.set_result(self)

            # Set the status watcher
            if self.watcher is not None:
                self.watcher.set()

    def finish(self, *args, **kwargs):
        """Convenience method to mark a command `~.CommandStatus.DONE`."""

        self.set_status(CommandStatus.DONE, *args, **kwargs)
        return self

    def fail(self, *args, **kwargs):
        """Convenience method to mark a command `~.CommandStatus.FAILED`."""

        self.set_status(CommandStatus.FAILED, *args, **kwargs)
        return self

    def debug(self, *args, **kwargs):
        """Writes a debug-level message."""

        self.write('d', *args, **kwargs)

    def info(self, *args, **kwargs):
        """Writes an info-level message."""

        self.write('i', *args, **kwargs)

    def warning(self, *args, **kwargs):
        """Writes a warning-level message."""

        self.write('w', *args, **kwargs)

    def error(self, *args, **kwargs):
        """Writes an error-level message (does not fail the command)."""

        self.write('e', *args, **kwargs)

    def write(self, message_code='i', message=None, broadcast=False, **kwargs):
        """Writes to the user(s).

        Parameters
        ----------
        message_code : str
            The message code (e.g., ``'i'`` or ``':'``).
        message : str or dict
            The text to be output. If `None`, only the code will be written.

        """

        if message is None:
            message = {}
        elif isinstance(message, dict):
            pass
        elif isinstance(message, str):
            message = {self.default_keyword: message}
        else:
            raise ValueError(f'invalid message {message!r}')

        if not self.actor:
            raise clu.CommandError('An actor has not been defined for '
                                   'this command. Cannot write to users.')

        result = self.actor.write(message_code, message=message, command=self,
                                  broadcast=broadcast, **kwargs)

        if asyncio.iscoroutine(result):
            self.loop.create_task(result)


class Command(BaseCommand):
    """A command from a user.

    Parameters
    ----------
    command_string : str
        The string that defines the body of the command.
    actor : ~clu.actor.BaseActor
        The actor instance associated to this command.
    transport
        The TCP transport associated with this command (only relevant
        for `.LegacyActor` commands).

    """

    def __init__(self, command_string='', actor=None, transport=None, **kwargs):

        BaseCommand.__init__(self, **kwargs)

        #: The raw command string.
        self.raw_command_string = command_string

        #: The body of the command.
        self.body = command_string

        #: The actor in which this command will run.
        self.actor = actor

        if not self.actor and kwargs.get('parent', None):
            self.actor = self.parent.actor

        #: The `~click.Context` running this command. Only relevant if
        #: using the built-in click-based parser.
        self.ctx = None

        self.transport = transport

    def parse(self):
        """Parses the command."""

        if not self.actor:
            raise clu.CluError('the actor is not defined. Cannot parse command.')

        self.actor.parse_command(self)

        return self

    def __str__(self):

        return (f'<Command (commander_id={self.commander_id!r}, '
                f'command_id={self.command_id!r}, body={self.body!r})>')


def parse_legacy_command(command_string):
    """Parses a command received by a legacy actor.

    Parameters
    ----------
    command_string : str
        The command to parse, including an optional header.

    Returns
    -------
    command_id, command_body : tuple
        The command ID, and the command body parsed from the command
        string.

    """

    _HEADER_BODY_RE = re.compile(r'((?P<cmdID>\d+)(?:\s+\d+)?\s+)?((?P<cmdBody>[A-Za-z_].*))?$')

    command_match = _HEADER_BODY_RE.match(command_string)
    if not command_match:
        raise clu.CommandError(f'Could not parse command {command_string!r}')

    command_dict = command_match.groupdict('')

    command_id = command_dict['cmdID']
    if command_id:
        command_id = int(command_id)
    else:
        command_id = 0

    command_body = command_dict.get('cmdBody', '').strip()

    return command_id, command_body



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2019-05-10
# @Filename: tron.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import asyncio

from clu.model import BaseModel, Property
from clu.protocol import TCPStreamClient

from .keys import KeysDictionary
from .parser import ParseError, ReplyParser


__all__ = ['TronConnection', 'TronModel', 'TronKey']


class TronKey(Property):
    """A Tron model key with callbacks.

    Differs from `.Property` in that it is usually instantiated by
    `.TronModel` with a ``key`` that is of type ``clu.legacy.keys.Key``
    instead of a string.

    """

    def __repr__(self):
        return f'<{self.__class__.__name__!s} ({self.key.name}): {self.value}>'


class TronModel(BaseModel):
    """A JSON-compliant model for actor keywords.

    Parameters
    ----------
    keydict : KeysDictionary
        A dictionary of keys that define the datamodel.
    callback
        A function or coroutine to call when the datamodel changes. The
        function is called with the instance of `TronModel` as the only
        parameter. If the callback is a coroutine, it is scheduled as a task.
    log : ~logging.Logger
        Where to log messages.

    """

    def __init__(self, keydict, callback=None, log=None):

        super().__init__(keydict.name, callback, log=log)

        self.keydict = keydict

        for key in self.keydict.keys:
            key = self.keydict.keys[key]
            self[key.name] = TronKey(key, model=self)

    def parse_reply(self, reply):
        """Parses a reply and updates the datamodel."""

        for reply_key in reply.keywords:

            key_name = reply_key.name.lower()
            if key_name not in self.keydict:
                if self.log:
                    self.log.warning('cannot parse unknown keyword '
                                     f'{self.name}.{reply_key.name}.')
                continue

            # When parsed the values in reply_key are string. After consuming
            # it with the Key, the values become typed values.
            result = self.keydict.keys[key_name].consume(reply_key)

            if not result:
                if self.log:
                    self.log.warning('failed parsing keyword '
                                     f'{self.name}.{reply_key.name}.')
                continue

            self[reply_key.name].value = [value.native for value in reply_key.values]

            if self.callback:
                self.scheduler.add_callback(self.callback, self)


class TronConnection(object):
    """Allows to send commands to Tron and manages the feed of replies.

    Parameters
    ----------
    actor : str
        The actor that is connecting to Tron. Used as the commander when
        sending commands through the hub.
    host : str
        The host on which Tron is running.
    port : int
        The port on which Tron is running.
    model_names : list
        A list of strings with the actors whose models will be tracked.
    log : ~logging.Logger
        Where to log messages.

    """

    def __init__(self, actor, host, port, model_names=None, log=None):

        self.actor = actor

        self.host = host
        self.port = port

        self.log = log

        self._mid = 1

        model_names = model_names or []

        #: dict: The `KeysDictionary` associated with each actor to track.
        self.keyword_dicts = {actor: KeysDictionary.load(actor)
                              for actor in model_names}

        #: dict: The `TronModel` instance holding the model and values of each actor being tracked.
        self.models = {actor: TronModel(self.keyword_dicts[actor], log=log)
                       for actor in model_names}

        self._parser = None

        self.connection = None

    async def start(self, get_keys=True):
        """Starts the connection to Tron.

        Parameters
        ----------
        get_keys : bool
            If `True`, gets all the keys in the models.

        """

        self.connection = await TCPStreamClient(self.host, self.port)

        self._parser = asyncio.create_task(self._parse_tron())

        if get_keys:
            asyncio.create_task(self.get_keys())

    def send_command(self, target, command_string, mid=None):
        """Sends a command through the hub.

        Parameters
        ----------
        target : str
            The actor to command.
        command_string : str
            The command to send.
        mid : int
            The message id. If `None`, a sequentially increasing value will
            be used. You should not specify a ``mid`` unless you really know
            what you're doing.

        """

        mid = mid or self._mid

        # The mid must be a 32-bit unsigned number.
        if mid >= 2**32:
            self._mid = mid = 1

        # The format for the SDSS Hub is "commander message_id target command"
        # where commander needs to start with a letter and have a program and
        # a user joined by a dot. Otherwise the command will be accepted but
        # the reply will fail to parse.
        command = f'{self.actor}.{self.actor} {mid} {target} {command_string}\n'

        self.connection.writer.write(command.encode())

        self._mid += 1

    async def get_keys(self):
        """Gets all the keys for the models being tracked."""

        # Number of keys to be requested at once
        n_keys = 10

        for model in self.models.values():

            actor = model.name
            keys = [key.lower() for key in model]

            for ii in range(0, len(keys), n_keys):

                keys_to_request = keys[ii:ii + n_keys]

                if len(keys_to_request) == 0:
                    break

                keys_joined = ' '.join(keys_to_request)

                command_string = f'getFor={actor} {keys_joined}'

                self.send_command('keys', command_string)

    async def _parse_tron(self):
        """Tracks new replies from Tron and updates the model."""

        rparser = ReplyParser()

        while True:

            line = await self.connection.reader.readline()

            try:
                line = line.decode()  # Do not strip here or that will cause parsing problems.
                reply = rparser.parse(line)
            except ParseError:
                if self.log:
                    self.log.debug(f'failed parsing reply {line}.')
                continue

            actor = reply.header.actor

            # The keys command returns keywords as if from the actor
            # keys_<actor> (e.g. keys_tcc).
            if actor.startswith('keys_'):
                actor = actor.split('_')[1]

            if actor not in self.models:
                continue

            try:
                self.models[actor].parse_reply(reply)
            except Exception as ee:
                if self.log:
                    self.log.debug(f'failed parsing reply {reply!r} with error: {ee!s}')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2019-10-05
# @Filename: testing.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import asyncio
import json
import re
import sys
import types
import unittest.mock

import clu
from clu.command import Command


if sys.version_info.major == 3 and sys.version_info.minor >= 8:
    CoroutineMock = unittest.mock.AsyncMock
else:
    try:
        from asynctest import CoroutineMock
    except ImportError:
        raise ImportError('clu.testing requires asynctest with Python < 3.8.')


__all__ = ['MockReply', 'MockReplyList', 'setup_test_actor', 'TestCommand']


class MockReply(dict):
    """Stores a reply written to a transport.

    The data of the message is stored as part of the dictionary.

    Attributes
    ----------
    user_id : int
        The user ID of the client to which the reply was sent.
    command_id : int
        The command ID of the command that produced this reply.
    flag : str
        The message type flag.
    data : dict
        The payload of the message.

    """

    def __init__(self, command_id, user_id, flag, data={}):

        self.command_id = command_id
        self.user_id = user_id
        self.flag = flag

        dict.__init__(self, data)

    def __repr__(self):
        return (f'<MockReply ({self.user_id} {self.command_id} '
                f'{self.flag} {super().__repr__()})>')


class MockReplyList(list):
    """Stores replies as `.MockReply` objects."""

    LEGACY_REPLY_PATTERN = re.compile(r'([0-9]+)\s+([0-9]+)\s+((?:[a-z]|\:))\s+(.*)')

    def __init__(self, actor):

        self.actor = actor

        list.__init__(self)

    def parse_reply(self, reply):
        """Parses a reply and construct a `.MockReply`, which is appended."""

        if isinstance(reply, bytes):
            reply = reply.decode()

        if issubclass(self.actor.__class__, clu.LegacyActor):

            match = self.LEGACY_REPLY_PATTERN.match(reply)
            if not match:
                return

            user_id, command_id, flag, keywords_raw = match.groups()

            user_id = int(user_id)
            command_id = int(command_id)

            data = {}
            for keyword_raw in keywords_raw.split(';'):
                if keyword_raw.strip() == '':
                    continue
                name, value = keyword_raw.split('=', maxsplit=1)
                data[name] = value

        elif issubclass(self.actor.__class__, clu.JSONActor):

            reply = json.loads(reply)

            header = reply['header']
            user_id = header.pop('commander_id', None)
            command_id = header.pop('command_id', None)
            flag = header.pop('message_code', 'd')

            data = reply['data']

        else:
            raise RuntimeError('actor must be LegacyActor or JSONActor.')

        list.append(self, MockReply(user_id, command_id, flag, data))

    def clear(self):
        list.__init__(self)

    def __contains__(self, m):

        return any([m in reply[kw] for reply in self for kw in reply.keys()])


async def setup_test_actor(actor, user_id=1):
    """Setups an actor for testing, mocking the client transport.

    Takes an ``actor`` and modifies it in two ways:

    - Adds a ``invoke_mock_command`` method to it that allows to submit a
      command string as if it had been received from a transport.

    - Mocks a client transport with ``user_id`` that is connected to the actor.
      Messages written to the transport are stored as `.MockReply` in a
      `.MockReplyList` that is accessible via a new ``actor.mock_replies``
      attribute.

    The actor is modified in place and also returned.

    """

    if not issubclass(actor.__class__, (clu.LegacyActor, clu.JSONActor)):
        raise RuntimeError('setup_test_actor is only usable with '
                           'LegacyActor or JSONActor actors.')

    def invoke_mock_command(self, command_str, command_id=0):
        if isinstance(command_str, str):
            command_str = command_str.encode('utf-8')
        full_command = f' {command_id} '.encode('utf-8') + command_str
        return self.new_command(actor.transports['mock_user'], full_command)

    actor.run = CoroutineMock(return_value=actor)

    # Adds an invoke_mock_command method.
    # We use types.MethodType to bind a method to an existing instance
    # (see http://bit.ly/35buk1m)
    actor.invoke_mock_command = types.MethodType(invoke_mock_command, actor)

    # Mocks a user transport and stores the replies in a MockReplyListobject
    actor.mock_replies = MockReplyList(actor)

    mock_transport = unittest.mock.MagicMock(spec=asyncio.Transport)
    mock_transport.user_id = user_id
    mock_transport.write.side_effect = actor.mock_replies.parse_reply

    actor.transports['mock_user'] = mock_transport

    actor = await actor.run()

    return actor


class TestCommand(Command):
    """A `.Command` that can be reset."""

    def reset(self):
        """Resets the command."""

        self._status = self.flags.READY

        if self.watcher:
            self.watcher.clear()
        self.watcher = None

        asyncio.Future.__init__(self, loop=self.loop)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2017-12-05
# @Filename: command.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)


__all__ = ['CluError', 'CluNotImplemented', 'CluWarning',
           'CluDeprecationWarning', 'CommandError', 'CommandParserError']


class CluError(Exception):
    """A custom core Clu exception"""

    def __init__(self, message=None):

        message = 'There has been an error' \
            if not message else message

        super(CluError, self).__init__(message)


class CluNotImplemented(CluError):
    """A custom exception for not yet implemented features."""

    def __init__(self, message=None):

        message = 'This feature is not implemented yet.' if not message else message

        super(CluNotImplemented, self).__init__(message)


class CommandError(CluError):
    """An error raised when a `Command` fails."""

    pass


class CommandParserError(CluError):
    """An error raised when parsing a command fails."""

    pass


class CluBaseWarning(Warning):
    """Base warning for Clu."""
    pass


class CluWarning(UserWarning, CluBaseWarning):
    """The primary warning class."""
    pass


class CluDeprecationWarning(CluBaseWarning):
    """A warning for deprecated features."""
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2019-05-06
# @Filename: parser.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import asyncio
import functools
import re

import click
from click.decorators import group, pass_obj

from . import actor


__all__ = ['CluCommand', 'CluGroup', 'command_parser',
           'ClickParser', 'timeout']


class CluCommand(click.Command):
    """Override `click.Command` to pass the actor and command as arguments."""

    def done_callback(self, task, exception_handler=None):
        """Checks if the command task has been successfully done."""

        log = getattr(task, 'log', None)
        ctx = task.ctx

        command = ctx.obj['parser_args'][0]

        if exception_handler and task.exception():
            exception_handler(command, task.exception(), log=log)

    async def _schedule_callback(self, ctx, timeout=None):
        """Schedules the callback as a task with a timeout."""

        if not hasattr(ctx, 'obj') or 'parser_args' not in ctx.obj:
            parser_args = []
            command = None
        else:
            parser_args = ctx.obj['parser_args']
            command = parser_args[0]

        callback_task = asyncio.create_task(
            ctx.invoke(self.callback, *parser_args, **ctx.params))

        try:
            await asyncio.wait_for(callback_task, timeout=timeout)
        except asyncio.TimeoutError:
            if command:
                command.set_status(command.status.TIMEDOUT,
                                   f'command timed out after {timeout} seconds.')
            return False

        return True

    def invoke(self, ctx):
        """Same as `click.Command.invoke` but passes the actor and command."""

        click.core._maybe_show_deprecated_notice(self)

        if self.callback is not None:

            with ctx:

                loop = asyncio.get_event_loop()

                # Makes sure the callback is a coroutine
                if not asyncio.iscoroutinefunction(self.callback):
                    self.callback = asyncio.coroutine(self.callback)

                # Check to see if there is a timeout value in the callback.
                # If so, schedules a task to be cancelled after timeout.
                timeout = getattr(self.callback, 'timeout', None)

                log = ctx.obj.pop('log', None)

                # Defines the done callback function.
                exception_handler = ctx.obj.pop('exception_handler', None)
                done_callback = functools.partial(self.done_callback,
                                                  exception_handler=exception_handler)

                # Launches callback scheduler and adds the done callback
                ctx.task = loop.create_task(self._schedule_callback(ctx, timeout=timeout))
                ctx.task.add_done_callback(done_callback)

                # Add some attributes to the task because it's
                # what will be passed to done_callback
                ctx.task.ctx = ctx
                ctx.task.log = log

            return ctx


class CluGroup(click.Group):
    """Override `click.Group` to make all child commands instances of `.CluCommand`."""

    def command(self, *args, **kwargs):
        """Override `click.Group` to use `.CluCommand` as class by default."""

        if 'cls' in kwargs:
            pass
        else:
            kwargs['cls'] = CluCommand

        def decorator(f):
            cmd = click.decorators.command(*args, **kwargs)(f)
            self.add_command(cmd)
            return cmd

        return decorator

    def parse_args(self, ctx, args):

        # Copy this method so that we can turn off the printing of the
        # usage before ctx.exit()
        if not args and self.no_args_is_help and not ctx.resilient_parsing:
            # click.echo(ctx.get_help(), color=ctx.color)
            ctx.exit()

        rest = click.Command.parse_args(self, ctx, args)
        if self.chain:
            ctx.protected_args = rest
            ctx.args = []
        elif rest:
            ctx.protected_args, ctx.args = rest[:1], rest[1:]

        return ctx.args

    def group(self, *args, **kwargs):
        """Creates a new group inheriting from this class."""

        if 'cls' not in kwargs:
            kwargs['cls'] = self.__class__

        def decorator(f):
            assert not asyncio.iscoroutinefunction(f), 'groups cannot be coroutines.'
            cmd = group(*args, **kwargs)(f)
            self.add_command(cmd)
            return cmd

        return decorator


def timeout(seconds):
    """A decorator to timeout the command after a number of ``seconds``."""

    def decorator(f):

        # This is a bit of a hack but we cannot access the context here so
        # we add the timeout directly to the callback function.
        f.timeout = seconds

        @functools.wraps(f)
        def new_func(*args, **kwargs):
            return f(*args, **kwargs)
        return functools.update_wrapper(new_func, f)

    return decorator


def pass_args():
    """Thing wrapper around pass_obj to pass the command and parser_args."""

    def decorator(f):
        @functools.wraps(f)
        @pass_obj
        def new_func(obj, *args, **kwargs):
            return f(*obj['parser_args'], *args, **kwargs)
        return functools.update_wrapper(new_func, f)

    return decorator


@click.group(cls=CluGroup)
def command_parser():
    pass


@command_parser.command()
def ping(*args):
    """Pings the actor."""

    command = args[0]
    command.set_status(command.status.DONE, 'Pong.')

    return


@command_parser.command(name='help')
@click.argument('PARSER-COMMAND', type=str, required=False)
@click.pass_context
def help_(ctx, *args, parser_command):
    """Shows the help."""

    command = args[0]

    # The parser_command arrives wrapped in quotes to make sure is a single
    # value. Strip it and unpack it in as many groups and commands as needed.
    parser_command = parser_command.strip('"').split()

    help_lines = ''

    # Gets the help lines for the command group or for a specific command.
    if len(parser_command) > 0:

        ctx_commands = ctx.command.commands

        for ii in range(len(parser_command)):
            ctx_command_name = parser_command[ii].lower()
            if ctx_command_name not in ctx_commands:
                return command.fail(text=f'command {ctx_command_name} not found.')
            ctx_command = ctx_commands[ctx_command_name]
            if ii == len(parser_command) - 1:
                # This is the last element in the command list
                # so we want to actually output this help lines.
                help_lines = ctx_command.get_help(ctx)
            else:
                ctx_commands = ctx_command.commands

    else:

        help_lines = ctx.get_help()

    message = []
    for line in help_lines.splitlines():
        # Remove the parser name.
        match = re.match(r'^Usage:([A-Za-z-\ ]+? \[OPTIONS\]) .*', line)
        if match:
            line = line.replace(match.groups()[0], '')

        message.append(line)

    if isinstance(command.actor, (actor.AMQPActor, actor.JSONActor)):
        return command.finish(help=message)
    else:
        for line in message:
            command.warning(help=line)
        return command.finish()


class ClickParser:
    """A command parser that uses Click at its base."""

    #: list: Arguments to be passed to each command in the parser.
    #: Note that the command is always passed first.
    parser_args = []

    parser = command_parser

    def parse_command(self, command):
        """Parses an user command using the Click internals."""

        # Makes sure we have the ping and help commands available
        if 'help' not in self.parser.commands:
            self.parser.add_command(help_)
        if 'ping' not in self.parser.commands:
            self.parser.add_command(ping)

        # This will pass the command as the first argument for each command.
        # If self.parser_args is defined, those arguments will be passed next.
        parser_args = [command]
        parser_args += self.parser_args

        # Empty command. Just finish the command.
        if not command.body:
            command.done()
            return command

        command.set_status(command.status.RUNNING)

        # If the command contains the --help flag,
        # redirects it to the help command.
        if '--help' in command.body:
            command.body = 'help ' + command.body
            command.body = command.body.replace(' --help', '')

        if not command.body.startswith('help'):
            command_args = command.body.split()
        else:
            command_args = ['help', '"{}"'.format(command.body[5:])]

        # We call the command with a custom context to get around
        # the default handling of exceptions in Click. This will force
        # exceptions to be raised instead of redirected to the stdout.
        # See http://click.palletsprojects.com/en/7.x/exceptions/
        ctx = self.parser.make_context(
            f'{self.name}-command-parser', command_args,
            obj={'parser_args': parser_args,
                 'log': self.log,
                 'exception_handler': self._handle_command_exception})

        # Makes sure this is the global context. This solves problems when
        # the actor have been started from inside an existing context,
        # for example when it's called from a CLI click application.
        click.globals.push_context(ctx)

        # Sets the context in the command.
        command.ctx = ctx

        with ctx:
            try:
                self.parser.invoke(ctx)
            except Exception as exc:
                self._handle_command_exception(command, exc)

        return command

    @staticmethod
    def _handle_command_exception(command, exception, log=None):
        """Handles an exception during parsing or execution of a command."""

        try:

            raise exception

        except (click.ClickException, click.exceptions.Exit) as ee:

            if not hasattr(ee, 'message'):
                ee.message = None

            ctx = command.ctx
            message = ''

            # If this is a command that cannot be parsed.
            if ee.message is None and ctx:
                message = f'{ee.__class__.__name__}:\n{ctx.get_help()}'
            else:
                message = f'{ee.__class__.__name__}: {ee.message}'

            lines = message.splitlines()
            for line in lines:
                command.write('w', text=line)

            msg = f'Command {command.body!r} failed.'

            if not command.status.is_done:
                command.fail(text=msg)
            else:
                command.write(text=msg)

        except click.exceptions.Exit:

            # This happens when using --help, although it should be handled
            # in parse_command.
            if command.status.is_done:
                command.write(text='Use help [CMD]')
            else:
                command.fail(text='Use help [CMD]')

        except click.exceptions.Abort:

            if not command.status.is_done:
                command.fail(text='Command was aborted.')

        except Exception:

            msg = (f'Command {command.command_id} failed because '
                   'of an uncaught error. See traceback in the log '
                   'for more information.')

            if command.status.is_done:
                command.write(text=msg)
            else:
                command.fail(text=msg)

            log = log or getattr(command.ctx, 'log', None)
            if log:
                log.exception(f'Command {command.body!r} failed with error:')



"""
Core classes for SDSS-3 keyword validation

Refer to https://trac.sdss3.org/wiki/Ops/Validation for details.
"""

# Created 18-Nov-2008 by David Kirkby (dkirkby@uci.edu)

import collections
import hashlib
import importlib
import sys
import textwrap

from . import html as utilHtml
from . import messages as protoMess
from . import types as protoTypes


class KeysError(Exception):
    pass


class Consumer(object):
    """
    Consumes parsed messages
    """

    # set debug True to generate detailed tracing of all the consume activity
    debug = False

    indent = 0

    def trace(self, what):
        if self.debug:
            print('%s%r << %r' % (' ' * Consumer.indent, self, what))
            Consumer.indent += 1

    def passed(self, what):
        if self.debug:
            Consumer.indent -= 1
            print('%sPASS >> %r' % (' ' * Consumer.indent, what))
        return True

    def failed(self, reason):
        if self.debug:
            Consumer.indent -= 1
            print('%sFAIL: %s' % (' ' * Consumer.indent, reason))
        return False

    def consume(self, what):
        raise NotImplementedError


class TypedValues(Consumer):
    """
    Consumes typed command or keyword values
    """

    def __init__(self, vtypes):
        self.vtypes = []
        self.minVals = 0
        self.maxVals = 0
        for vtype in vtypes:
            if isinstance(vtype, protoTypes.RepeatedValueType):
                if vtype.minRepeat != vtype.maxRepeat and vtype is not vtypes[-1]:
                    raise KeysError('Repetition range only allowed for last value type')
                self.minVals += vtype.minRepeat
                if self.maxVals is not None and vtype.maxRepeat is not None:
                    self.maxVals += vtype.maxRepeat
                else:
                    self.maxVals = None
                self.vtypes.append(vtype)
            elif isinstance(vtype, protoTypes.ValueType):
                self.minVals += 1
                self.maxVals += 1
                self.vtypes.append(vtype)
            elif isinstance(vtype, protoTypes.CompoundValueType):
                self.minVals += len(vtype.vtypes)
                self.maxVals += len(vtype.vtypes)
                self.vtypes.append(vtype)
            elif isinstance(vtype, protoTypes.ByName):
                self.vtypes.append(vtype)
            else:
                raise KeysError('Invalid value type: %r' % vtype)
        if self.maxVals == 0:
            self.descriptor = 'none'
        elif self.maxVals == self.minVals:
            self.descriptor = self.minVals
        elif self.maxVals is None:
            self.descriptor = '%d or more' % self.minVals
        else:
            self.descriptor = '%d-%d' % (self.minVals, self.maxVals)

    def __repr__(self):
        return 'Types%r' % self.vtypes

    def consume(self, values):
        self.trace(values)
        # remember the original values in case we need to restore them later
        self.originalValues = protoMess.Values(values)
        values[:] = []
        # try to convert each keyword to its expected type
        self.index = 0
        for typeToConsume in self.vtypes:
            if isinstance(typeToConsume, protoTypes.RepeatedValueType):
                vtype = typeToConsume.vtype
                offset = 0
                while offset < typeToConsume.minRepeat:
                    if not self.consumeNextValue(vtype, values):
                        values[:] = self.originalValues
                        return self.failed('expected repeated value type %r' % typeToConsume)
                    offset += 1
                while typeToConsume.maxRepeat is None or offset < typeToConsume.maxRepeat:
                    if not self.consumeNextValue(vtype, values):
                        break
                    offset += 1
            elif isinstance(typeToConsume, protoTypes.ValueType):
                if not self.consumeNextValue(typeToConsume, values):
                    values[:] = self.originalValues
                    return self.failed('expected value type %r' % typeToConsume)
            elif isinstance(typeToConsume, protoTypes.CompoundValueType):
                for vtype in typeToConsume.vtypes:
                    if not self.consumeNextValue(vtype, values):
                        values[:] = self.originalValues
                        return self.failed('expected compound value type %r' % typeToConsume)
                # Optionally replace the values with a reference to a single object
                # initialized with the values. The default object is a tuple.
                if protoTypes.CompoundValueType.WrapEnable:
                    size = len(typeToConsume.vtypes)
                    wrapped = tuple(values[-size:])
                    if typeToConsume.wrapper:
                        wrapped = typeToConsume.wrapper(*wrapped)
                    values[-size:] = [wrapped]
            else:
                raise KeysError('Unexpected typeToConsume: %r' % typeToConsume)
        if self.index != len(self.originalValues):
            values[:] = self.originalValues
            return self.failed('not all values consumed: %s' % values[self.index:])
        return self.passed(values)

    def consumeNextValue(self, valueType, values):
        try:
            string = self.originalValues[self.index]
            try:
                values.append(valueType(string))
            except protoTypes.InvalidValueError:
                values.append(protoTypes.InvalidValue)
            self.index += 1
            return True
        except (IndexError, ValueError, TypeError, OverflowError):
            return False

    def describeAsHTML(self):
        content = utilHtml.Div(
            utilHtml.Div(
                utilHtml.Span('Values', className='label'),
                utilHtml.Span(self.descriptor, className='value'),
                className='key descriptor'),
            className='vtypes')
        for vtype in self.vtypes:
            content.append(utilHtml.Div(utilHtml.Entity('nbsp'), className='separator'))
            content.extend(vtype.describeAsHTML().children)
        return content

    def describe(self):
        text = '%12s: %s\n' % ('Values', self.descriptor)
        for vtype in self.vtypes:
            extra = vtype.describe().replace('\n', '\n    ')
            text += '\n    %s\n' % extra
        return text


class Key(Consumer):
    """
    Base class for a command or reply keyword consumer

    Inputs:
    - name: keyword name
    - a list of one or more value types
    - help: an optional help string
    - doCache: is it reasonable to read this value from a cache?
        defaults to True only if the keyword can have values and you do not specify refreshCmd
    - refreshCmd: an optional command that can be sent to the actor to refresh this value
    """

    def __init__(self, name, *vtypes, **metadata):
        self.name = name
        self.typedValues = TypedValues(vtypes)
        self.help = metadata.get('help', None)
        self.refreshCmd = metadata.get('refreshCmd', None)
        defDoCache = (self.typedValues.maxVals != 0) and not self.refreshCmd
        self.doCache = bool(metadata.get('doCache', defDoCache))
        if 'unique' in metadata:
            self.unique = metadata.get('unique')

    def __repr__(self):
        return 'Key(%s)' % self.name

    def consume(self, keyword):
        self.trace(keyword)
        # perform a case-insensitive name matching
        if keyword.name.lower() != self.name.lower():
            return self.failed('keyword has wrong name')
        if not self.typedValues.consume(keyword.values):
            return self.failed('no match for keyword values')
        keyword.matched = True
        return self.passed(keyword)

    def create(self, *values):
        """
        Returns a new Keyword using this Key as a template

        Any keyword values must match the expected types or else this
        method returns None. The returned keyword will have typed
        values.
        """
        if len(values) == 1 and isinstance(values[0], list):
            values = values[0]
        keyword = protoMess.Keyword(self.name, values)
        if not self.consume(keyword):
            raise KeysError('value types do not match for keyword %s: %r' % (self.name, values))
        return keyword

    def describeAsHTML(self):
        content = utilHtml.Div(utilHtml.Div(self.name, className='keyname'), className='key')
        desc = utilHtml.Div(className='keydesc')
        content.append(desc)
        if self.help:
            desc.append(
                utilHtml.Div(
                    utilHtml.Span('Description', className='label'),
                    utilHtml.Span(self.help, className='value'),
                    className='key descriptor'))
        desc.extend(self.typedValues.describeAsHTML().children)
        return content

    def describe(self):
        text = '%12s: %s\n' % ('Keyword', self.name)
        if self.help:
            pad = '\n' + ' ' * 14
            formatted = textwrap.fill(
                textwrap.dedent(self.help).strip(), width=66).replace('\n', pad)
            text += '%12s: %s\n' % ('Description', formatted)
        text += self.typedValues.describe()
        return text


class KeysManager(object):

    keys = {}

    # These need to be classmethods, not static methods, so that
    # subclasses each access their own keys list.
    @classmethod
    def setKeys(cls, kdict):
        cls.keys = {}
        cls.addKeys(kdict)

    @classmethod
    def addKeys(cls, kdict):
        if not isinstance(kdict, KeysDictionary):
            raise KeysError('Cmd keys must be provided as a KeysDictionary')
        cls.keys[kdict.name] = kdict

    @classmethod
    def getKey(cls, name):
        for kdict in cls.keys.values():
            if name in kdict:
                return kdict[name]
        raise KeysError('No such registered keyword <%s>' % name)


class CmdKey(Consumer, KeysManager):
    """
    Consumes a command keyword
    """

    def __init__(self, key):
        self.key = key

    def __repr__(self):
        return 'CmdKey(%s)' % self.key.name

    def consume(self, where):
        self.trace(where)
        keyword = where.keyword()
        if not keyword:
            return self.failed('no keywords available to consume')
        if not self.key.consume(keyword):
            return self.failed('no match for command keyword')
        where.advance()
        return self.passed(where)


class RawKey(Consumer):
    """
    Consumes the special 'raw' keyword in a command
    """

    def consume(self, where):
        self.trace(where)
        keyword = where.keyword()
        if not keyword:
            return self.failed('no keywords available to consume')
        if not isinstance(keyword, protoMess.RawKeyword):
            return self.failed('no match for raw keyword')
        # no re-casting as a typed value needed here
        where.advance()
        return self.passed(where)


class KeysDictionaryError(KeysError):
    pass


class KeysDictionary(object):
    """
    A collection of Keys associated with a given name

    The dictionary name is typically the name of an actor. Contains a
    registry of all known KeysDictionaries, for use by the load method.
    """
    registry = {}

    def __init__(self, name, version, *keys):
        """
        Creates a new named keys dictionary

        Overwrites any existing dictionary with the same name. The
        version should be specified as a (major,minor) tuple of
        integers. Dictionary names must be lower case.
        """
        self.name = name
        try:
            (major, minor) = map(int, version)
        except (ValueError, TypeError):
            raise KeysDictionaryError(
                'Invalid version: expected (major,minor) tuple of integers, got %r' % version)
        self.version = version
        if not name == name.lower():
            raise KeysDictionaryError('Invalid name: must be lower case: %s' % name)
        KeysDictionary.registry[name] = self
        self.keys = collections.OrderedDict()
        self.namedTypes = {}
        for key in keys:
            self.add(key)

    def add(self, key):
        """
        Adds a key to the dictionary

        By default, the key is registered under key.name but this can be
        overridden by setting a key.unique attribute. Attempting to
        add a key using a name that is already assigned raises an
        exception.
        """
        if not isinstance(key, Key):
            raise KeysDictionaryError('KeysDictionary can only contain Keys')
        name = getattr(key, 'unique', key.name)
        if name.lower() in self.keys:
            raise KeysDictionaryError('KeysDictionary name is not unique: %s' % name)
        # look for named types used by the key
        localNames = {}
        for index, vtype in enumerate(key.typedValues.vtypes):
            vname = getattr(vtype, 'name', None)
            if isinstance(vtype, protoTypes.ByName):
                try:
                    key.typedValues.vtypes[index] = self.namedTypes[vname]
                except KeyError:
                    raise KeysDictionaryError('Unresolved type ByName("%s")' % vname)
            elif vname:
                localNames[vname] = vtype
        # add this key's local names to our dictionary (ByName can only refer to
        # types defined in a previously-defined type to ensure unique names)
        self.namedTypes.update(localNames)
        # add this key
        self.keys[name.lower()] = key

    def extend(self, keyList):
        """Adds all the keys in a list."""

        assert isinstance(keyList, (list, tuple)), 'keyList must be a list or tuple'

        for key in keyList:
            self.add(key)

    def __getitem__(self, name):
        return self.keys[name.lower()]

    def __contains__(self, name):
        return name.lower() in self.keys

    def describe(self):
        """
        Generates text describing all of our keys in alphabetical order
        """
        text = 'Keys Dictionary for "%s" version %r\n' % (self.name, self.version)
        for name in sorted(self.keys):
            text += '\n' + self.keys[name].describe()
        return text

    def describeAsHTML(self):
        """
        Generates HTML describing all of our keys in alphabetical order
        """
        content = utilHtml.Div(
            utilHtml.Div(
                utilHtml.Span(self.name, className='actorName'),
                utilHtml.Span('%d.%d' % self.version, className='actorVersion'),
                className='actorHeader'),
            className='actor')
        for name in sorted(self.keys):
            content.append(self.keys[name].describeAsHTML())
        return content

    @staticmethod
    def load(dictname, forceReload=False):
        """
        Loads a KeysDictionary by name, returning the result

        Uses an in-memory copy, if one is available, otherwise loads the
        dictionary from disk. Use forceReload to force the dictionary to
        be loaded from disk even if it is already in memory. Raises a
        KeysDictionaryError if a dictionary cannot be found for dictname.
        """
        if not forceReload and dictname in KeysDictionary.registry:
            return KeysDictionary.registry[dictname]
        # try to find a corresponding file on the import search path
        dictfile = None
        try:
            # get the path corresponding to the actorkeys package
            import actorkeys  # noqa
            keyspath = sys.modules['actorkeys'].__path__
        except ImportError:
            raise KeysDictionaryError('no actorkeys package found')
        try:
            # open the file corresponding to the requested keys dictionary
            (dictfile, name, description) = importlib.util.find_spec(dictname, keyspath)
            # create a global symbol table for evaluating the keys dictionary expression
            symbols = {
                '__builtins__': __builtins__,
                'Key': Key,
                'KeysDictionary': KeysDictionary,
                'ByName': protoTypes.ByName,
            }
            for (name, value) in protoTypes.__dict__.items():
                if isinstance(value, type) and issubclass(
                        value, (protoTypes.ValueType, protoTypes.CompoundValueType)):
                    symbols[name] = value
            # evaluate the keys dictionary as a python expression
            filedata = dictfile.read()
            kdict = eval(filedata, symbols)
            # check that the dictionary filename and name match
            if not dictname == kdict.name:
                raise KeysDictionaryError(
                    'dictionary filename and name are different: %s, %s' % (dictname, kdict.name))
            # do a checksum so that we can detect changes independently of versioning
            kdict.checksum = hashlib.md5(filedata.encode()).hexdigest()
            return kdict
        except ImportError as e:
            raise KeysDictionaryError('no keys dictionary found for %s: %s' % (dictname, str(e)))
        except Exception as e:
            indent = '\n >> '
            description = indent + indent.join(str(e).split('\n'))
            raise KeysDictionaryError(
                'badly formatted keys dictionary in %s:%s' % (dictfile.name, description))
        finally:
            if dictfile:
                dictfile.close()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'python'}

packages = \
['clu', 'clu.legacy', 'clu.legacy.ply']

package_data = \
{'': ['*']}

install_requires = \
['Click>=7.0,<8.0',
 'aio_pika>=6.4.1,<7.0.0',
 'jsonschema>=3.0.1,<4.0.0',
 'sdsstools>=0.2.0']

extras_require = \
{'docs': ['Sphinx>=2.3']}

setup_kwargs = {
    'name': 'sdss-clu',
    'version': '0.3.1',
    'description': 'A new protocol for SDSS actors.',
    'long_description': "`CLU <https://tron.fandom.com/wiki/Clu>`__\n==========================================\n\n|py| |Build Status| |docs|\n\n\n`CLU <https://tron.fandom.com/wiki/Clu>`_ implements a new protocol for SDSS actor while providing support for legacy-style actor.\n\n\nFeatures\n--------\n\n- Asynchronous API based on `asyncio <https://docs.python.org/3/library/asyncio.html>`_.\n- New-style actor with message passing based on `AMQP <https://www.amqp.org/>`_ and `RabbitMQ <https://rabbitmq.com>`_.\n- Legacy-style actor for TCP socket communication through ``Tron``.\n- Tools for device handling.\n- Messages are validated JSON strings.\n- `click <https://click.palletsprojects.com/en/7.x/>`__-enabled command parser.\n\n.. warning:: CLU is under active development and it must be considered in beta stage. The API is changing quickly and breaking changes are frequent.\n\n\nInstallation\n------------\n\n``CLU`` can be installed using ``pip`` as ::\n\n    pip install sdss-clu\n\nor from source ::\n\n    git clone https://github.org/sdss/clu\n    cd clu\n    pip install .\n\nIf you want to install it using `modules <http://modules.sourceforge.net/>`_ you can use `sdss_install <https://github.com/sdss/sdss_install>`_ ::\n\n    sdss_install clu\n\n\nQuick start\n-----------\n\nCreating a new actor with ``CLU`` is easy. To instantiate and run an actor you can simply do ::\n\n    import asyncio\n    from clu import AMQPActor\n\n    async def main(loop):\n        actor = Actor('guest', 'localhost', loop=loop).run()\n\n    loop = asyncio.get_event_loop()\n    loop.create_task(main(loop))\n    loop.run_forever()\n\nNext, head to the `Getting started <https://clu.readthedocs.io/en/latest/getting-started.html>`_ section for more information about using actors.\n\n\nWhy a new messaging protocol for SDSS?\n--------------------------------------\n\nSay whatever you want about it, the `current SDSS message passing protocol <https://clu.readthedocs.io/en/latest/legacy.html>`_ based on ``Tron``, ``opscore``, and ``actorcore`` is stable and robust. So, why should we replace it? Here is a list of reasons:\n\n- It reinvents the wheel. Ok, in all honesty ``Tron`` and ``opscore`` were written when wheel were still not completely circular, but the truth is that nowadays there are more robust, standard, and better documented technologies out there for message passing.\n- We can remove the need for a central hub product by relying in open-source message brokers such as `RabbitMQ <https://rabbitmq.com>`__.\n- ``Tron`` and ``opscore`` are Python 2 and it's unclear the amount of effort that would be needed to convert them to Python 3.\n- While there is some documentation for ``Tron`` and ``opscore``, and the code is well written, it's also cumbersome and difficult to modify by people that didn't write it. It's ultimately non-maintainable.\n- The ``opsctore``/``actorkeys`` datamodel is custom-built and extremely difficult to maintain. Standard solutions such as JSON with a `JSON schema <https://json-schema.org/>`__ validator should be preferred.\n- `asyncio <https://docs.python.org/3/library/asyncio.html>`__ provides an asynchronous API that is cleaner and easier to code than using threads. It is also more readable and less convoluted than `twisted <https://twistedmatrix.com/trac/>`__ and it's a Python core library with very active development.\n- CLU uses `click <https://click.palletsprojects.com/en/7.x>`__ for parsing commands, providing a well-defined, easy to use parser.\n\n\n.. |Build Status| image:: https://img.shields.io/github/workflow/status/sdss/drift/Test\n    :alt: Build Status\n    :target: https://github.com/sdss/clu/actions\n\n.. |Coverage Status| image:: https://codecov.io/gh/sdss/clu/branch/master/graph/badge.svg\n    :alt: Coverage Status\n    :scale: 100%\n    :target: https://codecov.io/gh/sdss/clu\n\n.. |py| image:: https://img.shields.io/badge/python-3.7%20|%203.8-blue\n    :alt: Python Versions\n    :target: https://docs.python.org/3/\n\n.. |docs| image:: https://readthedocs.org/projects/docs/badge/?version=latest\n    :alt: Documentation Status\n    :scale: 100%\n    :target: https://clu.readthedocs.io/en/latest/?badge=latest\n",
    'author': 'José Sánchez-Gallego',
    'author_email': 'gallegoj@uw.edu',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/sdss/clu',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2019-05-14
# @Filename: device.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import asyncio
import contextlib

from .protocol import TCPStreamClient
from .tools import CallbackScheduler


__all__ = ['Device']


class Device(object):
    """A class that handles the TCP connection to a device.

    There are two ways to create a new device. You can create a subclass from
    `.Device` and override the `.process_message` method which handles how you
    react to a new line being received ::

        class MyDevice(Device):

            async def process_message(self, line):
                print(line)

        my_device = MyDevice('192.168.1.10', 4444)
        await my_device.start()

    Note that `.process_message` must be a coroutine. Alternatively you can
    pass a callback that will be called instead of `.process_message` when
    a new message arrives. The callback must also be a coroutine ::

        async def printer(line):
            print(line)

        my_device = MyDevice('192.168.1.10', 4444, callback=printer)
        await my_device.start()


    Parameters
    ----------
    host : str
        The host of the device.
    port : int
        The port on which the device is serving.
    callback
        The coroutine to call when a new message is received. The coroutine
        gets a single argument with all the buffer received from the client
        until a newline arrives. If no callback is specified,
        `.process_message` is called. The callback is always awaited and it is
        the user's responsibility to handle long tasks appropriately. If the
        callback is not a coroutine, it will be converted to one.

    """

    def __init__(self, host, port, callback=None):

        self.host = host
        self.port = port

        #: TCPStreamClientContainer: the connection to the device.
        self.connection = None
        self.listener = None
        self.scheduler = CallbackScheduler()

        self.callback = callback or self.process_message
        if not asyncio.iscoroutinefunction(self.callback):
            self.callback = asyncio.coroutine(self.callback)

    async def start(self):
        """Opens the connection and starts the listener."""

        if self.is_connected():
            raise RuntimeError('connection is already running.')

        self.connection = await TCPStreamClient(self.host, self.port)
        self.listener = asyncio.create_task(self._listen())

    async def stop(self):
        """Closes the connection and stops the listener."""

        self.connection.close()
        await self.connection.writer.wait_closed()  # Waits until it's really closed.

        with contextlib.suppress(asyncio.CancelledError):
            self.listener.cancel()
            await self.listener

    def is_connected(self):
        """Returns `True` if the connection is open."""

        if self.connection is None:
            return False

        return not self.connection.writer.is_closing()

    def write(self, message, newline='\n'):
        """Write to the device. The message is encoded and a new line added."""

        assert self.is_connected() and self.connection.writer, 'device is not connected'

        message = message.strip() + newline
        self.connection.writer.write(message.encode())

    async def _listen(self):
        """Listens to the reader stream and callbacks on message received."""

        if not self.connection:
            raise RuntimeError('connection is not open.')

        while True:
            line = await self.connection.reader.readline()
            line = line.decode().strip()
            await self.callback(line)

    async def process_message(self, line):
        """Processes a newly received message."""

        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2018-01-16
# @Filename: actor.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import asyncio
import json
import re
import time
import uuid
from contextlib import suppress

from .base import BaseActor
from .client import AMQPClient
from .command import Command
from .exceptions import CommandError
from .model import ModelSet
from .parser import ClickParser
from .protocol import TCPStreamServer
from .tools import log_reply


try:
    import aio_pika as apika
except ImportError:
    apika = None


__all__ = ['AMQPActor', 'JSONActor', 'TimerCommand', 'TimerCommandList']


class AMQPActor(AMQPClient, ClickParser, BaseActor):
    """An actor class that uses AMQP message brokering.

    This class differs from `~clu.legacy.actor.LegacyActor` in that it uses
    an AMQP messaging broker (typically RabbitMQ) to communicate with other
    actors in the system, instead of standard TCP sockets. Although the
    internals and protocols are different the entry points and behaviour for
    both classes should be almost identical.

    See the documentation for `.AMQPClient` for parameter information.

    """

    def __init__(self, *args, model_path=None, model_names=None, **kwargs):

        AMQPClient.__init__(self, *args, **kwargs)

        self.commands_queue = None

        if model_path:

            model_names = model_names or []

            if self.name not in model_names:
                model_names.append(self.name)

            self.models = ModelSet(model_path, model_names=model_names,
                                   raise_exception=False, log=self.log)

        else:

            self.log.warning('no models loaded.')
            self.models = None

        self.timer_commands = TimerCommandList(self)

    async def start(self, **kwargs):
        """Starts the connection to the AMQP broker."""

        # This sets the replies queue but not a commands one.
        await AMQPClient.start(self, **kwargs)

        # Binds the commands queue.
        self.commands_queue = await self.connection.add_queue(
            f'{self.name}_commands', callback=self.new_command,
            bindings=[f'command.{self.name}.#'])

        self.log.info(f'commands queue {self.commands_queue.name!r} '
                      f'bound to {self.connection.connection.url!s}')

        self.timer_commands.start()

        return self

    async def new_command(self, message):
        """Handles a new command received by the actor."""

        with message.process():

            headers = message.info()['headers']
            command_body = json.loads(message.body.decode())

        commander_id = headers['commander_id'].decode()
        command_id = headers['command_id'].decode()
        command_string = command_body['command_string']

        try:
            command = Command(command_string, command_id=command_id,
                              commander_id=commander_id,
                              consumer_id=self.name,
                              actor=self, loop=self.loop)
            command.actor = self  # Assign the actor
        except CommandError as ee:
            await self.write('f', {'text': f'Could not parse the '
                                           f'following as a command: {ee!r}'})
            return

        return self.parse_command(command)

    async def handle_reply(self, message):
        """Handles a reply received by the message and updates the models.

        Parameters
        ----------
        message : aio_pika.IncomingMessage
            The message received.

        """

        reply = await AMQPClient.handle_reply(self, message)

        if self.models and reply.sender in self.models:
            self.models[reply.sender].update_model(reply.body)

        return

    async def write(self, message_code='i', message=None, command=None,
                    broadcast=False, **kwargs):
        """Writes a message to user(s).

        Parameters
        ----------
        message_code : str
            The message code (e.g., ``'i'`` or ``':'``).
        message : dict
            The keywords to be output. Must be a dictionary of pairs
            ``{keyword: value}``.
        command : Command
            The command to which we are replying. If not set, it is assumed
            that this is a broadcast.
        broadcast : bool
            Whether to broadcast the message to all the users or only to the
            commander.
        kwargs
            Keyword arguments that will be added to the message.

        """

        message = message or {}

        assert isinstance(message, dict), 'message must be a dictionary'

        message.update(kwargs)

        message_json = json.dumps(message)

        if command is None:
            broadcast = True

        commander_id = command.commander_id if command else None
        command_id = command.command_id if command else None

        headers = {'message_code': message_code,
                   'commander_id': commander_id,
                   'command_id': command_id,
                   'sender': self.name}

        if broadcast:
            routing_key = 'reply.broadcast'
        else:
            routing_key = f'reply.{command.commander_id}'

        await self.connection.exchange.publish(
            apika.Message(message_json.encode(),
                          content_type='text/json',
                          headers=headers,
                          correlation_id=command_id),
            routing_key=routing_key)

        log_reply(self.log, message_code, message_json)


class JSONActor(ClickParser, BaseActor):
    """A TCP actor that replies using JSON.

    This implementation of `.BaseActor` uses TCP as command/reply channel and
    replies to the user by sending a JSON-valid string. This makes it useful
    as a "device" actor that is not connected to the central message parsing
    system but that we still want to accept commands and reply with easily
    parseable messages.

    Commands received by this actor must be in the format
    ``[<uid>] <command string>``, where ``<uid>`` is any integer unique
    identifier that will be used as ``command_id`` and appended to any reply.

    Parameters
    ----------
    name : str
        The actor name.
    host : str
        The host where the TCP server will run.
    port : int
        The port of the TCP server.
    args,kwargs
        Arguments to be passed to `.BaseActor`.

    """

    def __init__(self, name, host, port, *args, **kwargs):

        super().__init__(name, *args, **kwargs)

        self.host = host
        self.port = port

        #: Mapping of commander_id to transport
        self.transports = dict()

        #: TCPStreamServer: The server to talk to this actor.
        self.server = TCPStreamServer(host, port, loop=self.loop,
                                      connection_callback=self.new_user,
                                      data_received_callback=self.new_command)

        self.timer_commands = TimerCommandList(self)

    async def start(self):
        """Starts the TCP server."""

        await self.server.start_server()
        self.log.info(f'running TCP server on {self.host}:{self.port}')

        self.timer_commands.start()

        return self

    async def run_forever(self):
        """Runs the actor forever, keeping the loop alive."""

        await self.server.serve_forever()

    def new_user(self, transport):
        """Assigns userID to new client connection."""

        if transport.is_closing():
            if hasattr(transport, 'user_id'):
                self.log.debug(f'user {transport.user_id} disconnected.')
                return self.transports.pop(transport.user_id)

        user_id = str(uuid.uuid4())
        transport.user_id = user_id
        self.transports[user_id] = transport

        sock = transport.get_extra_info('socket')
        if sock is not None:
            peername = sock.getpeername()[0]
            self.log.debug(f'user {user_id} connected from {peername}.')

        return

    def new_command(self, transport, command_str):
        """Handles a new command received by the actor."""

        commander_id = getattr(transport, 'user_id', None)
        message = command_str.decode().strip()

        if not message:
            return

        command_id, command_string = re.match(r'([0-9]*)\s*(.+)', message).groups()

        if command_id == '':
            command_id = 0
        else:
            command_id = int(command_id)

        command_string = command_string.strip()

        if not command_string:
            return
        try:
            command = Command(command_string=command_string,
                              commander_id=commander_id,
                              command_id=command_id,
                              consumer_id=self.name,
                              actor=self, loop=self.loop,
                              transport=transport)
        except CommandError as ee:
            self.write('f', {'text': f'Could not parse the following as a command: {ee!r}'})
            return

        return self.parse_command(command)

    def send_command(self):
        """Not implemented for `.JSONActor`."""

        raise NotImplementedError('JSONActor cannot send commands to other actors.')

    def write(self, message_code='i', message=None, command=None,
              broadcast=False, beautify=True, **kwargs):
        """Writes a message to user(s) as a JSON.

        A ``header`` keyword with the ``commander_id`` (i.e., the user id of
        the transport that sent the command), ``command_id``, ``message_code``,
        and ``sender`` is added to each message. The payload of the message
        is written to ``data``. An example of a valid message is

        .. code-block:: yaml

            {
                "header": {
                    "command_id": 0,
                    "commander_id": 1,
                    "message_code": "i",
                    "sender": "test_camera"
                },
                "message": {
                    "camera": {
                        "name": "test_camera",
                        "uid": "DEV_12345"
                    },
                    "status": {
                        "temperature": 25.0,
                        "cooler": 10.0
                    }
                }
            }

        Parameters
        ----------
        message_code : str
            The message code (e.g., ``'i'`` or ``':'``). Ignored.
        message : dict
            The keywords to be output. Must be a dictionary of pairs
            ``{keyword: value}``.
        command : Command
            The command to which we are replying. If not set, it is assumed
            that this is a broadcast.
        broadcast : bool
            Whether to broadcast the message to all the users or only to the
            commander.
        beautify : bool
            Whether to format the JSON to make it more readable.
        kwargs
            Keyword arguments that will used to update the message.

        """

        message = message or {}
        assert isinstance(message, dict), 'message must be a dictionary'
        message.update(kwargs)

        commander_id = command.commander_id if command else None
        command_id = command.command_id if command else None
        transport = command.transport if command else None

        message_full = {}
        header = {'header': {'command_id': command_id,
                             'commander_id': commander_id,
                             'message_code': message_code,
                             'sender': self.name}
                  }

        message_full.update(header)
        message_full.update({'data': message})

        if beautify:
            message_json = json.dumps(message_full, sort_keys=False, indent=4)
        else:
            message_json = json.dumps(message_full, sort_keys=False)

        message_json += '\n'

        if broadcast or commander_id is None or transport is None:
            for transport in self.transports.values():
                transport.write(message_json.encode())
        else:
            transport.write(message_json.encode())

        log_reply(self.log, message_code, message_json)


class TimerCommandList(list):
    """A list of `.TimerCommand` objects that will be executed on a loop.

    Parameters
    ----------
    actor
        The actor in which the commands are to be run.
    resolution : float
        In seconds, how frequently to check if any of the `.TimerCommand` must
        be executed.

    """

    def __init__(self, actor, resolution=0.5, loop=None):

        self.resolution = resolution
        self.actor = actor
        self._task = None
        self.loop = loop or asyncio.get_event_loop()

        list.__init__(self, [])

    def add_command(self, command_string, **kwargs):
        """Adds a new `.TimerCommand`."""

        self.append(TimerCommand(command_string, **kwargs))

    async def poller(self):
        """The polling loop."""

        current_time = time.time()

        while True:

            for timer_command in self:
                elapsed = current_time - timer_command.last_run
                if elapsed > timer_command.delay:
                    timer_command_task = self.loop.create_task(timer_command.run(self.actor))
                    timer_command_task.add_done_callback(timer_command.done)

            self._sleep_task = self.loop.create_task(asyncio.sleep(self.resolution))

            await self._sleep_task
            current_time += self.resolution

    def start(self):
        """Starts the loop."""

        if self.running:
            raise RuntimeError('poller is already running.')

        self._task = self.loop.create_task(self.poller())

        return self

    async def stop(self):
        """Cancel the poller."""

        if not self.running:
            return

        self._task.cancel()

        with suppress(asyncio.CancelledError):
            await self._task

    @property
    def running(self):
        """Returns `True` if the poller is running."""

        if self._task and not self._task.cancelled():
            return True

        return False


class TimerCommand(object):
    """A command to be executed on a loop.

    Parameters
    ----------
    command_string : str
        The command string to run.
    delay : float
        How many seconds to wait between repeated calls.

    """

    def __init__(self, command_string, delay=1):

        self.command_string = command_string
        self.delay = delay

        self.last_run = 0.0

    async def run(self, actor):
        """Run the command."""

        await Command(self.command_string, actor=actor).parse()

    def done(self, task):
        """Marks the execution of a command."""

        self.last_run = time.time()
